(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)
(** Linear Arithmetique *)

(**
   This module use one domain and one semantic value.
   They are both polynomes, but the domain can be reduced to a class
   (p = 1 * cl + 0) but the semantic value can't.
*)
open Popop
open Stdlib
open Types
open Solver

let debug = Debug.register_info_flag
  ~desc:"for the arithmetic theory"
  "arith"

let debug_todo = debug


open Polynome
let dom_invariant = Polynome.invariant
let sem_invariant p = Polynome.is_one_cl p == None && dom_invariant p


let sem : t sem = Sem.create_key "SARITH"
let dom : t dom = Dom.create_key "ARITH"
let real_ctr = Ty.Constr.create "Real"
let real = Ty.ctr real_ctr

let minus_or_one inv =
  if inv then Q.minus_one else Q.one

let sub_cl cl p =
  x_p_cy (monome Q.one cl) Q.minus_one p

(** cl1 -> cl2 *)
let dist cl1 cl2 =
  add (monome Q.one cl1) (monome Q.minus_one cl2)

(** Real summary
   merge cl1 cl2 -> cl1 - cl2 = 0
   set_sem cl1 sem v -> cl1 - v = 0
   dist cl1 cl2 -> cl1 - cl2 = 0
*)

(** A merge gives two thing, a knowledge for the pivot
    and a knowledge for the two classes merged *)
(*
type expkind =
| ExpSolve of pexp * Q.t * Cl.t option * Cl.t option
(** pexp, div, cl_o, cl_r


    Since when P, P_o, P_r have been decided cl_r = cl_o, then
    dom(cl_r) = dom(cl_o) and we propagated for the pivot q that
    (dom(cl_r) - dom(cl_o) - div q)/div = q

    (q is a pivot, coef_o * q \in dom(cl_o) and coef_r * q \in dom(cl_r)
    div = coef_o + coef_r <> 0 but coef_o or coef_r can be null)

    With the following explications:
    pexp -> +- P + cl_r = cl_o
    cl_r  -> P_r + dom(cl_r) = cl_r
    cl_o  -> P_o + dom(cl_o) = cl_o

    the explication of this propagation is: (+- P + P_r - P_o)/div
    because
    (+- P + P_r - P_o)/div + (dom(cl_r) - dom(cl_o) - div q)/div = q

    +- depend on the non-deterministic choice of the solver for the equivalence
    class for cl_r and cl_o.
*)
*)
type expequality =
(* | ExpSem of expkind * Cl.t *)
(* (\** ExpSem (exp,q): propagate a sem: exp = q *\) *)
(* | ExpCl of expkind * Cl.t * Cl.t *)
(* (\** ExpCl (exp,p,q): propagate a merge: exp + p = q *\) *)
(* | ExpSubst of Cl.t * norm *)
| ExpCombi  of int * t * combi (** int only used for debug *)
(** All the class have been simplified
  remains a constant different than zero *)
| ExpCstDiff of Q.t * combi
| ExpMult of Q.t (**p**) * Cl.t (**g**) * Cl.t (**f**) * Cl.t (**own**)
     (** p = g -> g*f = p*f **)

(** linear combination *)
and combi =
| CombiRepr of Q.t * Cl.t * combi
| CombiDom of Q.t * Cl.t * combi
| CombiSem of Q.t * t * combi
| CombiPexp of Q.t * t * Explanation.pexp * combi
| CombiNull

let rec iter_combi f = function
  | CombiNull -> ()
  | CombiRepr (_,_,n)
  | CombiDom (_,_,n) | CombiPexp (_,_,_,n)
  | CombiSem (_,_,n) as c ->
    f c; iter_combi f n

let rec mult_combi m = function
  | CombiNull -> CombiNull
  | CombiRepr (q,a1,n) -> CombiRepr(Q.mul m q,a1,mult_combi m n)
  | CombiDom (q,a1,n) -> CombiDom(Q.mul m q,a1,mult_combi m n)
  | CombiPexp (q,th,a1,n) ->
    CombiPexp(Q.mul m q,th,a1,mult_combi m n)
  | CombiSem (q,a1,n) -> CombiSem(Q.mul m q,a1,mult_combi m n)

(** Substitution daemon *)

module T = struct
  include Polynome
  let key = sem
end

module ThE = RegisterSem(T)

let propa_dem = Demon.Key.create "Arith.DaemonPropa"

let expequality : expequality Explanation.exp =
  Explanation.Exp.create_key "Arith.equality"

let print_bag_cl = Bag.print Pp.comma Cl.print

let attach d clsem b1 =
  Debug.dprintf4 debug "[Arith] @[attach@ %a@ for@ %a@]@\n"
    ThE.print clsem print_bag_cl b1;
  let open Demon.Key in
  match is_attached d propa_dem clsem with
  | SDead | SRedirected _ ->
    Debug.dprintf2 debug "[Arith] @[Dead or redirected! of %a@]@\n"
      ThE.print clsem;
    assert false (** absurd: it was considered not
                     useful anymore it can't be needed now *)
  | SAlive b2 ->
    Demon.Key.set_info d propa_dem clsem (Bag.concat b1 b2)
  | SUnborn ->
    Debug.dprintf2 debug "[Arith] @[attach propagation of %a@]@\n"
      ThE.print clsem;
    let events = Polynome.fold (fun acc cl _ ->
        Demon.Create.EventDom(cl,dom,cl)::
        begin if Delayed.get_dom d dom cl == None
          then Demon.Create.EventChange(cl,cl)::acc
          else acc
        end
    ) [] (ThE.sem clsem) in
    Demon.Key.attach d propa_dem clsem events;
    Demon.Key.set_info d propa_dem clsem b1

module Th = struct
  include T

  let merged v1 v2 =
    match v1,v2 with
    | None, None -> true
    | Some v', Some v -> T.equal v' v
    | _ -> false

    type 'a normalizeResult =
    | NormalizeTo of Cl.t
    | NormalizeToSem of 'a

    let norm x =
      match is_one_cl x with
      | None -> NormalizeToSem x
      | Some cl -> NormalizeTo cl

    let subst_sp pi x = function
      | NormalizeTo y ->
        subst_cl pi x y
      | NormalizeToSem s ->
        subst pi x s

    let normalize d v =
      Polynome.fold
        (fun (norm,p) x _ ->
          let x' = Delayed.find_def d x in
          match Delayed.get_dom d dom x' with
          | None when Cl.equal x x' -> norm,p
          | None   -> let p,q = subst_cl p x x' in
            (if Q.equal q Q.zero then norm else CombiRepr(q,x,norm)), p
          | Some s -> let p,q = subst p x s in
            (if  Q.equal q Q.zero then norm else CombiDom(q,x,norm)), p)
         (CombiNull,v) v

    let filter_to_update d cls =
    (** filter out the classes that are the same can append because
          two classes wainting for their merge have the same dom, and
          this dom can be be exactly the same so no merge happend for
          this dom at the end *)
      if Bag.is_num_elt 1 cls then cls
      else
        let m = Bag.fold_left
            (fun m cl -> Cl.M.add (Delayed.find d cl) cl m) Cl.M.empty cls in
        Bag.list (Cl.M.values m)

    (* let equal_sr sr1 sr2 = *)
    (*   match sr1, sr2 with *)
    (*   | NormalizeTo cl1, NormalizeTo cl2 -> Cl.equal cl1 cl2 *)
    (*   | NormalizeToSem p1, NormalizeToSem p2 -> equal p1 p2 *)
    (*   | _ -> false *)

    (** combi: q*x + p' = 0 *)
    let apply_pivot d combi q x p' =
        (** sr = (-1/q) * p' *)
        let iq = Q.inv q in
        let miq = Q.neg iq in
        let s = mult_cst miq p' in
        let sr = norm s in
        let expkind = mult_combi miq combi in
        let miq_p' = mult_cst miq p' in
        let mx = monome Q.one x in
        let clsemmx = ThE.index mx real in
        let cls,nodom =
          match Demon.Key.is_attached d propa_dem clsemmx with
          | SDead | SRedirected _ ->
            assert false (** absurd: it was considered not
                     useful anymore it can't be needed now *)
          | SAlive b ->
            Debug.dprintf2 debug "@[b = %a@]@\n" print_bag_cl b;
            assert (Bag.exists (fun y -> Delayed.is_equal d x y) b);
            b, false
          | SUnborn -> Bag.elt x, true
        in
        let cls = filter_to_update d cls in
        Demon.Key.kill d propa_dem clsemmx;
        let age = Delayed.current_age d in
        let pexp own =
          if nodom
          then begin
            assert (Cl.equal own x);
            Delayed.mk_pexp ~age d expequality
              (ExpCombi(1,sub_cl own miq_p',expkind))
          end
          else
            Delayed.mk_pexp ~age d expequality
              (ExpCombi(15,sub_cl own miq_p',CombiDom(Q.one,own,expkind)))
        in
        begin match sr with
          | NormalizeTo y ->
            Debug.dprintf6 debug
              "[Arith] @[merge q=%a x=%a@, y=%a@]@\n"
              Q.pp_print q Cl.print x Cl.print y;
            Delayed.merge d (pexp x) x y;
            let subst = monome Q.one y in
            let clsemsubst = ThE.index subst real in
            let iter own =
              let pexp = pexp own in
              Delayed.set_dom d pexp dom own subst;
            in
            Bag.iter iter cls;
            let cls =
              match Delayed.get_dom d dom y with
              | None ->
                let combinull = Delayed.mk_pexp d expequality
                    (ExpCombi(3,Polynome.zero,CombiNull)) in
                Delayed.set_dom d combinull dom y subst;
                Bag.add y cls
              | Some s' ->
                assert (T.equal subst s');
                cls
            in
            attach d clsemsubst cls
          | NormalizeToSem subst ->
            Debug.dprintf6 debug
              "[Arith] @[set_sem q=%a x=%a@, s=%a@]@\n"
              Q.pp_print q Cl.print x print subst;
            let clsemsubst = ThE.index subst real in
            let iter own =
              let pexp = pexp own in
              Delayed.set_dom d pexp dom own subst
            in
            Bag.iter iter cls;
            let clsem = ThE.clsem (ThE.index subst real) in
            Delayed.set_sem d (pexp x) x clsem;
            attach d clsemsubst cls
        end


    (** p = p1 - p2 *)
    let solve_aux d combi p =
      match Polynome.extract p with
      | Zero -> () (** nothing to do already equal *)
      | Cst q ->
        let pexp = Delayed.mk_pexp d expequality
            (ExpCstDiff(q, combi)) in
        Delayed.contradiction d pexp
      | Var(q,x,p') ->
        (** p = qx + p' *)
        assert ( not (Q.equal Q.zero q) );
        Debug.dprintf2 debug "[Arith] @[solve p=%a@]@\n" print p;
        apply_pivot d combi q x p'

    let norm_dom d cl = function
      | None ->
        let r = Delayed.find d cl in
        let r = monome Q.one r in
        let clsemr = ThE.index r real in
        let pexp = Delayed.mk_pexp d expequality
            (ExpCombi(2,sub_cl cl r,
                      CombiRepr(Q.one,cl,CombiNull))) in
        Delayed.set_dom d pexp dom cl r;
        attach d clsemr (Bag.elt cl);
        r
      | Some p ->
        p


    let merge d pexp_merge (p1o,cl1) (p2o,cl2) _ =
      assert (not (Delayed.is_equal d cl1 cl2));
      assert (not (p1o == None && p2o == None));
      let p1 = norm_dom d cl1 p1o in
      let p2 = norm_dom d cl2 p2o in
      let dist = dist cl1 cl2 in (* pexp *)
      let diff,diff_pexp = sub p1 p2,
                           CombiPexp(Q.minus_one,dist, pexp_merge,
                           CombiDom(Q.one,cl1,
                           CombiDom(Q.minus_one,cl2,
                           CombiNull))) in
      solve_aux d diff_pexp diff

    let merge_itself d combi cl1 p1 =

      let p2o = Delayed.get_dom d dom cl1 in
      match p2o with
      | Some p2 when equal p1 p2 -> ()
      | _ ->
      let p2 = norm_dom d cl1 p2o in
      let pexp = CombiDom(Q.minus_one,cl1,combi) in
      let diff,diff_pexp = sub p1 p2, pexp in
      solve_aux d diff_pexp diff


end

module D = RegisterDom(struct include Th let key = dom end)

let embed t cl =
  match Delayed.get_dom t dom cl with
  | None   -> monome Q.one cl
  | Some r -> r

module DaemonPropa = struct
  let key = propa_dem

  module Key = ThE
  module Data = Cl
  type info = Cl.t Bag.t let default = Bag.empty

  let immediate = true
  let wakeup d clsem ev cls =
    let v = ThE.sem clsem in
    let cls = Th.filter_to_update d cls in
    let norm, v' = List.fold_left (fun acc ev ->
        let f (norm,p) x =
          match Delayed.get_dom d dom x with
          | None ->
            let x' = Delayed.find d x in
            let p,q = subst_cl p x x' in
            (if Q.equal Q.zero q then norm else CombiRepr(q,x,norm)), p
          | Some v ->
            let p,q = subst p x v in
            (if Q.equal Q.zero q
             then norm
             else CombiDom(q,x,norm)), p
        in
        match ev with
        | Events.Fired.EventChange(_,cl) ->
          Debug.dprintf4 debug "Cl: x:%a p:%a@\n"
            Cl.print cl Th.print (snd acc);
          f acc cl
        | Events.Fired.EventDom(_,dom',cl) ->
          Debug.dprintf4 debug "Dom: x:%a p:%a@\n"
            Cl.print cl Th.print (snd acc);
          assert (Dom.equal dom dom');
          f acc cl
          | _ -> raise UnwaitedEvent
      ) (CombiNull,v) ev in
        Debug.dprintf4 debug "[Arith] @[subst %a@ into %a@]@\n"
          Th.print v Th.print v';
        assert (Th.equal v' (snd (Th.normalize d v)));
    let clsem' = ThE.index v' real in
    if ThE.equal clsem' clsem then Demon.AliveReattached
    else
    let pexp own =
      Delayed.mk_pexp d expequality
        (ExpCombi(8,sub_cl own v',
                  (CombiDom(Q.one,own,norm)))) in
        begin match Th.norm v' with
        | Th.NormalizeTo y    ->
        let iter own =
          let pexp = pexp own in
          Delayed.set_dom d pexp dom own v';
          Delayed.merge d pexp own y
        in
        Bag.iter iter cls;
        attach d clsem' (Bag.add y cls)
        | Th.NormalizeToSem s ->
        let clsem = ThE.clsem (ThE.index s real) in
        let iter own =
          let pexp = pexp own in
          Delayed.set_dom d pexp dom own s;
          Delayed.set_sem d pexp own clsem
        in
        Bag.iter iter cls;
        attach d clsem' cls
        end;
        Demon.AliveStopped
end

module RDaemonPropa = Demon.Key.Register(DaemonPropa)

module DaemonInit = struct
  let key = Demon.Key.create "Arith.DaemonInit"

  module Key = DUnit
  module Data = DUnit
  type info = unit let default = ()

  let immediate = false

  let wakeup d () ev () =
    List.iter
      (function Events.Fired.EventRegSem(clsem,()) ->
        begin
          let clsem = ThE.coerce_clsem clsem in
          let v = ThE.sem clsem in
          let own0 = ThE.cl clsem in
          match Demon.Key.is_attached d DaemonPropa.key clsem with
          | SDead | SAlive _ | SRedirected _ -> ()
          | SUnborn ->
            assert (sem_invariant v);
            Polynome.iter (fun x _ -> Delayed.register d x) v;
            let norm,v' = Th.normalize d v in
            let norm = CombiSem(Q.one,v,norm) in
            Debug.dprintf4 debug "[Arith] @[normalize %a@ into %a@]@\n"
              Th.print v Th.print v';
            Th.merge_itself d norm own0 v'
        end;
        Delayed.flush d;
      | _ -> raise UnwaitedEvent
      ) ev;
    Demon.AliveReattached

end

module RDaemonInit = Demon.Key.Register(DaemonInit)

(** For the multiplication *)
type mul = Mul of Cl.t * Cl.t
let make_mul f g = if Cl.compare f g > 0 then Mul(f,g) else Mul(g,f)

let mul_sem : mul sem = Sem.create_key "Arith.(*)"

type expmulsubst = mul
let expmulsubst : expmulsubst Explanation.exp =
  Explanation.Exp.create_key "Arith.(*)"


module Mul = struct
  module T = struct
    type t = mul
    let equal n m =
      match n, m with
      | Mul (g1,f1), Mul(g2,f2) -> Cl.equal g1 g2 && Cl.equal f1 f2

    let hash n =
      match n with
      | Mul(g,f) -> 3 * (Cl.hash g) + 5 * (Cl.hash f)

    let compare n m =
      match n, m with
      | Mul(g1,f1), Mul(g2,f2) ->
        let c = Cl.compare g1 g2 in
        if c <> 0 then c
        else Cl.compare f1 f2

    let print fmt v =
      match v with
      | Mul (f,g) -> Format.fprintf fmt "(@[%a@] * @[%a@])"
                       Cl.print  f Cl.print g

  end

  include MkDatatype(T)

  let key = mul_sem

  let cl_ord f g = if Cl.compare f g > 0 then f,g else g,f
end

module MulE = RegisterSem(Mul)


let norm_mul d own = function
  | Mul(f0,g0) as v ->
    let f = Delayed.find d f0 in
    let g = Delayed.find d g0 in
    let res = match Opt.bind is_cst (Delayed.get_dom d dom f) with
      | Some q ->
        Some (monome q g0, ExpMult(q,f0,g0,own))
      | _ ->
        match Opt.bind is_cst (Delayed.get_dom d dom g) with
        | Some q ->
          Some (monome q f0, ExpMult(q,g0,f0,own))
        | _ -> None in
    match res with
    | None when Cl.equal f f0 && Cl.equal g g0 ->
      Demon.AliveReattached
    | None (** congruence closure *) ->
      let pexp = Delayed.mk_pexp d expmulsubst v in
      let f,g = Mul.cl_ord f g in
      let clsem = MulE.clsem (MulE.index (Mul(f,g)) real) in
      Delayed.set_sem d pexp own clsem;
      Demon.AliveStopped
    | Some (r,exp) ->
      let pexp = Delayed.mk_pexp d expequality exp in
      begin match Th.norm r with
        | Th.NormalizeTo y    ->
          Delayed.merge d pexp own y
        | Th.NormalizeToSem s ->
          let clsem = ThE.clsem (ThE.index s real) in
          Delayed.set_sem d pexp own clsem
      end;
      Demon.AliveStopped



module DaemonMultPropa = struct
  let key = Demon.Key.create "Uninterp.DaemonMultPropa"

  module Key = Mul
  module Data = DUnit
  type info = unit let default = ()

  let immediate = false
  let wakeup d v _ev () =
    let own = Cl.index mul_sem v real in
      Debug.dprintf4 debug "[Uninterp] @[wakeup own %a v:%a@]@\n"
        Cl.print own Mul.print v;
      norm_mul d own v
end

module RDaemonMultPropa = Demon.Key.Register(DaemonMultPropa)

module DaemonMultInit = struct
  type k = unit
  type d = unit
  let key = Demon.Key.create "Uninterp.DaemonMultInit"

  module Key = DUnit
  module Data = DUnit
  type info = unit let default = ()

  let immediate = false

(*
  let _close_assoc d x y own =
    let use = Delayed.are_waiting_cl_change d x DaemonMultPropa.key in
    List.iter (fun ((Mul(z1,z2) as m),()) ->
    match Delayed.is_sem_present d mul_sem m with
    | None -> assert false
    | Some other ->
      let z = if Cl.equal z1 x then z2 else (assert (Cl.equal x z2); z1) in
      let newcl = Delayed.index d mul_sem (make_mul other y) in
      Delayed.register d newcl;
      Delayed.set_sem d mul_sem newcl (make_mul own z);
    ) use
*)

  let wakeup d () ev () =
    List.iter
      (function Events.Fired.EventRegSem(clsem,()) ->
        begin
          let clsem = MulE.coerce_clsem clsem in
          let v = MulE.sem clsem in
          let own = MulE.cl clsem in
          match v with
          | Mul(f,g) ->
            Debug.dprintf2 debug "[Arith] @[init mult %a@]@\n" Mul.print v;
            Delayed.register d f; Delayed.register d g;
            match norm_mul d own v with
            | AliveReattached ->
              (* close_assoc d f g own; *)
              (* close_assoc d g f own; *)
              Demon.Key.attach d DaemonMultPropa.key v
                [Demon.Create.EventDom(f,dom,());
                 Demon.Create.EventDom(g,dom,());
                 Demon.Create.EventChange(f,());
                 Demon.Create.EventChange(g,());
                ]
            | AliveStopped | AliveRedirected _ -> ()
        end
      | _ -> raise UnwaitedEvent
      ) ev;
    Demon.AliveReattached

end

module RDaemonMultInit = Demon.Key.Register(DaemonMultInit)

let zero = ThE.cl (ThE.index Polynome.zero real)

let th_register env =
  RDaemonPropa.init env;
  RDaemonInit.init env;
  RDaemonMultPropa.init env;
  RDaemonMultInit.init env;
  Demon.Key.attach env
    DaemonInit.key () [Demon.Create.EventRegSem(sem,())];
  Demon.Key.attach env
    DaemonMultInit.key () [Demon.Create.EventRegSem(mul_sem,())];
  Delayed.register env zero


(** Conflict *)
(** Reason of equalities between arithmetical terms *)
type conpoly = {imp : Th.t; exp : Th.t}
let conpoly : conpoly Explanation.con = Explanation.Con.create_key "Arith.eq"

let mk_conpoly p = {imp = p; exp = p}
let zero_conpoly = {imp = Polynome.zero; exp = Polynome.zero}

let add_conpoly p1 p2 =
  if p2 == zero_conpoly then p1
  else if p1 == zero_conpoly then p2
  else
  {imp = add p1.imp p2.imp; exp = add p1.exp p2.exp}

let x_p_cy_conpoly p1 q p2 =
  if p2 == zero_conpoly then p1
  else
    {imp = x_p_cy p1.imp q p2.imp; exp = x_p_cy p1.exp q p2.exp}

let get_pexp_conpoly_deps t pexp deps =
  let open Conflict in
  let v, deps' = ComputeConflict.get_pexp_deps t pexp conpoly in
  let deps = Explanation.Deps.concat deps' deps in
  match v with
  | GRequested p -> Some p, deps
  | GOther(con,c) ->
    None, Explanation.Deps.add_unknown_con deps con c

let get_pexp_conpoly t pexp =
  let r, deps = get_pexp_conpoly_deps t pexp Explanation.Deps.empty in
  Conflict.ComputeConflict.add_deps t deps;
  r

let dist_conpoly cl1 cl2 =
  let g = dist cl1 cl2 in
  {imp=g;exp=g}

let print_conpoly fmt t =
  Format.fprintf fmt "{imp=%a;exp=%a}" Th.print t.imp Th.print t.exp

let get_rlist_conpoly_deps t age cl1 cl2 deps =
  let rlist = Conflict.ComputeConflict.get_equal t age cl1 cl2 in
  Debug.dprintf6 debug "[Arith] get_rlist_conpoly cl1:%a cl2:%a: %a@\n"
    Cl.print cl1 Cl.print cl2 Conflict.print_rlist rlist;
  let r,deps,cl2' = List.fold_left
    (fun (p,deps,cl) edge ->
      let r,deps = get_pexp_conpoly_deps t edge.Conflict.epexp deps in
      let p = match r with
        | None -> add_conpoly p (dist_conpoly cl edge.Conflict.eto)
        | Some p' ->
          Debug.dprintf6 debug
            "[Arith] get_rlist_conpoly %a..%a: Done %a@\n"
            Cl.print cl Cl.print edge.Conflict.eto print_conpoly p';
          assert (Th.equal p'.exp (if edge.Conflict.einv
                                   then dist edge.Conflict.eto cl
                                   else dist cl edge.Conflict.eto));
          x_p_cy_conpoly p (minus_or_one edge.Conflict.einv) p'
      in
       p, deps, edge.Conflict.eto)
    (zero_conpoly,deps,cl1)
    rlist in
  Debug.dprintf6 debug "[Arith] get_rlist_conpoly %a -> %a: Done %a@\n"
    Cl.print cl1 Cl.print cl2' print_conpoly r;
  assert (Cl.equal cl2 cl2');
  assert (Th.equal r.exp (dist cl1 cl2));
  r,deps

let get_rlist_conpoly t age cl1 cl2 =
  let r, deps = get_rlist_conpoly_deps t age cl1 cl2 Explanation.Deps.empty in
  Conflict.ComputeConflict.add_deps t deps;
  r

(** Gen Equality and disequality *)
module GenEquality = struct
  open Conflict
  open Explanation

  let equality t age cl1 cl2 =
    (** cl1 -> cl2 *)
    let p = get_rlist_conpoly t age cl1 cl2 in
    (** cl2 -> cl1 *)
    let p = add_conpoly p (dist_conpoly cl2 cl1) in
    (** cl1 -> cl2 -> cl1 = 0 *)
    assert (is_zero p.exp);
    ComputeConflict.unknown_con t conpoly p

  let expspecial =
    let f = fun (b1,deps) _ b2 -> add_conpoly b1 b2, deps in
    let get_sem t deps age sem v pexp =
      let b,deps = get_pexp_conpoly_deps t pexp deps in
      raise TODO
      (* Explanation.Deps.add_unknown_con deps conpoly b *)
    in
    { Equality.equality = equality;
      disequality = (fun t age ~hyp cl1d cl1e cl2e cl2d ->
          equality t age cl1d cl1e;
          equality t age cl2d cl2e);
      merged = (fun t deps age cl1d cl1 pexp cl2 cl2d ->
          Debug.dprintf8 (* TODO *) debug
            "cl1d:%a cl1:%a cl2:%a cl2d:%a@\n"
            Cl.print cl1d Cl.print cl1
            Cl.print cl2  Cl.print cl2d;
          (** cl1d -> cl1 *)
          let pr1,deps = get_rlist_conpoly_deps t age cl1d cl1 deps in
          Debug.dprintf2 debug "cl1d -> cl1: %a@\n" print_conpoly pr1;
          assert (Th.equal pr1.exp (dist cl1d cl1));
          (** cl1 -> cl2 *)
          let ppexp, deps = get_pexp_conpoly_deps t pexp deps in
          let ppexp = match ppexp with
            | None -> dist_conpoly cl1 cl2
            | Some p -> p in
          Debug.dprintf2 debug "cl1 -> cl2: %a@\n" print_conpoly ppexp;
          assert (Th.equal ppexp.exp (dist cl1 cl2));
          (** cl2 -> cl2d *)
          let pr2,deps = get_rlist_conpoly_deps t age cl2 cl2d deps in
          Debug.dprintf2 debug "cl2 -> cl2d: %a@\n" print_conpoly pr2;
          assert (Th.equal pr2.exp (dist cl2 cl2d));
          (** cl2d -> cl1d *)
          let pd = dist_conpoly cl2d cl1d in
          Debug.dprintf2 debug "cl2d -> cl1d: %a@\n" print_conpoly pd;
          (** cl1d -> cl1 -> cl2 -> cl2d -> cl1d = 0 *)
          let p = add_conpoly (add_conpoly (add_conpoly pr1 ppexp) pr2) pd in
          Debug.dprintf2 debug "sum: %a@\n" print_conpoly p;
          assert (is_zero p.exp);
          Explanation.Deps.add_unknown_con deps conpoly p);
      repr = (fun t deps age cl r1 -> assert false);
          (* let p,deps = get_rlist_conpoly_deps t cl r1 deps in *)
          (* assert (is_zero p.exp); *)
          (* Explanation.Deps.add_unknown_con deps conpoly p); *)
      dodec = true (** TODO *);
    }

  let () = Equality.register_sort real expspecial

end

module ConPoly = struct
  open Explanation
  open Conflict
  type t = conpoly

  let print = print_conpoly

  let key = conpoly

  class finalized v : Conflict.finalized = object
    method print fmt =
      Pp.print_iter1 Th.S.iter Pp.semi
        (fun fmt x -> Format.fprintf fmt "%a=0" Th.print x) fmt v
    method test d =
      try
        Th.S.fold_left (fun acc v -> (** different than 0 *)
            let _,vn = Th.normalize d v in
            match Polynome.classify vn with
            | ZERO -> acc
            | CST -> raise Exit
            | VAR -> ToDecide
          ) False v
      with Exit -> True
    method decide :
      'a. 'a Conflict.fold_decisions -> Solver.Delayed.t -> 'a -> 'a =
      fun f d acc ->
        Th.S.fold_left (fun acc v ->
            let _,vn = Th.normalize d v in
            match Polynome.is_cst vn with
            | Some _ -> acc
            | None ->
              let f cl =
                if Equality.is_equal d cl zero ||
                   Equality.is_disequal d cl zero then acc
                else
                  let c = Equality.Eq(cl,zero,false) in
                  f.fold_decisions acc Equality.choequal c ()
              in
              match Th.is_one_cl v with
              | Some cl -> f cl
              | None -> f (ThE.cl (ThE.index v real))
          ) acc v
    method conflict_add d =
      let fold acc v =
        let f cl1 cl2 =
          let eq = Equality.disequality [cl1;cl2] in
          Cl.M.add eq false acc in
        (* let _,v = Th.normalize d v in *)
        (** todo: why the explication can be ignored? *)
        match Polynome.is_cst v with
        | Some _ -> acc
        | None ->
          (** no find because the explication would be different *)
          match Polynome.is_one_cl v with
          | Some cl' -> f cl' zero
          | None ->
            let cl' = Cl.index sem v real in
            f cl' zero
      in
      Th.S.fold_left fold Cl.M.empty v

  end

    (* let print fmt sl = *)
    (*   Bag.iter (fun p -> *)
    (*       Format.fprintf fmt "@[{imp=%a|exp=%a}@];" *)
    (*         Th.print p.imp Th.print p.exp) sl *)
    (* in *)


  let finalize _ sl =
    Debug.dprintf2 debug "[Arith] @[sl:%a@]@\n"
      (Bag.print Pp.semi print_conpoly) sl;
    let s = Bag.fold_left (fun acc p ->
        assert (is_zero p.exp);
        if is_zero p.imp then acc else Th.S.add p.imp acc
      ) Th.S.empty sl in
    let print = Pp.print_iter1 Th.S.iter Pp.semi Th.print in
    Debug.dprintf2 Conflict.print_conflicts
      "[Arith] @[conflict:%a.@]@\n" print s;
    if Th.S.is_empty s
    then None
    else Some (new finalized s)


  let same_sem (type a) t age (sem':a sem) (v:a) pexp1 cl1 cl2 =
    let r1 = get_pexp_conpoly t pexp1 in
    let p = match r1 with
      | Some p1 ->
        let p2 =
          match Sem.Eq.eq_type sem sem' with
          | None -> raise Impossible (* understand why that happend *)
          | Some Types.Eq ->
            let p2 = x_p_cy (monome Q.one cl2) Q.minus_one v in
            {imp = p2; exp = p2}
        in
        x_p_cy_conpoly p1 Q.minus_one p2
      | _ -> dist_conpoly cl1 cl2 in
    Debug.dprintf6 debug_todo "@[same_sem cl1:%a cl2:%a = %a@]@\n"
      Cl.print cl1 Cl.print cl2 print_conpoly p;
    assert (Th.equal p.exp (dist cl1 cl2));
    GRequested p


  let propacl t age cl rcl =
    if ComputeConflict.before_first_dec t age
    then GRequested (dist_conpoly cl rcl)
    else GRequested {imp = Polynome.zero; exp = dist cl rcl}

end

module EConPoly = Conflict.RegisterCon(ConPoly)

module ExpEquality = struct
  open Conflict
  (* open IterExp *)
  open ComputeConflict

  type t = expequality
(*
  let print_kind fmt = function
    |  ExpSolve (pexp,q,cl1,cl2) ->
      Format.fprintf fmt "ExpSolve(%a,%a,%a,%a)"
        print_pexpclcl pexp
        Q.pp_print q
        (Pp.print_option Cl.print) cl1
        (Pp.print_option Cl.print) cl2
    |  ExpMerge (pexp,q1,q2,cl1,cl2) ->
      Format.fprintf fmt "ExpMerge(%a,%a,%a,%a,%a)"
        print_pexpclcl pexp
        Q.pp_print q1
        Q.pp_print q2
        (Pp.print_option Cl.print) cl1
        (Pp.print_option Cl.print) cl2
*)

  let print_combi1 fmt = function
    | CombiRepr (q,cl,_) -> Format.fprintf fmt "%a~%a" Q.pp_print q Cl.print cl
    | CombiDom  (q,cl,_) -> Format.fprintf fmt "%a.%a" Q.pp_print q Cl.print cl
    | CombiSem    (q,t,_) -> Format.fprintf fmt "%a:%a" Q.pp_print q Th.print t
    | CombiPexp (q,def,pexp,_) -> Format.fprintf fmt "%a~(%a)%a"
                                Q.pp_print q Th.print def
                                Conflict.print_pexp pexp
    | CombiNull          -> Format.pp_print_string fmt "0"

  let print_combi fmt n =
    Pp.print_iter1 iter_combi (Pp.constant_string "+") print_combi1 fmt n

  let print fmt = function
(*    |  ExpSem (kind,cl) ->
      Format.fprintf fmt "ExpSem(%a,%a)"
        print_kind kind
        Cl.print cl
    |  ExpCl (kind,cl1,cl2) ->
      Format.fprintf fmt "ExpCl(%a,%a,%a)"
        print_kind kind
        Cl.print cl1
        Cl.print cl2
    | ExpSubst (cl,_) -> Format.fprintf fmt "Subst(%a)" Cl.print cl
*)
    | ExpCombi  (id,t,n)  -> Format.fprintf fmt "Combi(%i,%a,%a)"
                               id Th.print t print_combi n
    | ExpCstDiff (q,n) ->
      Format.fprintf fmt "ExpCstDiff(%a,%a)"
        Q.pp_print q
        print_combi n
    | ExpMult (p,g,f,own) ->
      Format.fprintf fmt "ExpMult(%a,%a,%a,%a)"
        Q.pp_print p Cl.print g Cl.print f Cl.print own

(*
  let iterexp t age x =
    let iter t age (pexp,clo,clr) cl1 cl2 =
        need_cl_repr t age clo;
        need_cl_repr t age clr;
        need_pexp t pexp;
        Opt.iter (fun x -> need_dom t age x dom) cl1;
        Opt.iter (fun x -> need_dom t age x dom) cl2 in
    let rec iternorm t age = function
      | SubstCl(_,cl,n)  -> need_cl_repr t age cl; iternorm t age n
      | SubstSem(_,cl,n) -> need_dom t age cl dom; iternorm t age n
      | SubstNil -> () in
    let iter_kind t age = function
      | ExpSolve(pexp,_,cl1,cl2)
      | ExpMerge(pexp,_,_,cl1,cl2) ->
        iter t age pexp cl1 cl2 in
    match x with
    | ExpSem(kind,_)
    | ExpCl(kind,_,_)  ->
      iter_kind t age kind
    | ExpSubst (cl,n) ->
      need_dom t age cl dom;
      iternorm t age n
    | ExpNorm  (v,n) ->
      need_sem t age sem v;
      iternorm t age n
    | ExpCstDiff(pexp,cl1,cl2) ->
      iter t age pexp cl1 cl2
*)


  let rec analyse_combi t age combi =
    Debug.dprintf2 debug_todo "@[analyse_combi? %a@]@\n"
      print_combi1 combi;
    let p = match combi with
    | CombiNull -> zero_conpoly
    | CombiPexp(q,th,pexp,n) ->
      let n = analyse_combi t age n in
      x_p_cy_conpoly n q
        (match get_pexp_conpoly t pexp with
         | None -> {imp=th;exp=th}
         | Some p -> assert (Th.equal p.exp th); p)
    | CombiDom(q,cl,n) ->
      let n = analyse_combi t age n in
      let mod_dom = get_dom_last t age cl dom in
      let p = x_p_cy_conpoly n
          (** dom -> modcl *)
          q (Opt.get (get_pexp_conpoly t mod_dom.Explanation.modpexp)) in
      let modcl = mod_dom.Explanation.modcl in
      (** cl -> modcl *)
      x_p_cy_conpoly p q (get_rlist_conpoly t age cl modcl)
    | CombiSem(q,v,n) ->
      let n = analyse_combi t age n in
      let cl = ThE.cl (ThE.index v real) in
      let pexp = mk_conpoly (x_p_cy (monome Q.one cl) Q.minus_one v) in
      x_p_cy_conpoly n q pexp
    | CombiRepr(q,cl,n) ->
      let n = analyse_combi t age n in
      let rcl = ComputeConflict.get_repr_at t age cl in
      let p = get_rlist_conpoly t age cl rcl in
      assert (Th.equal p.exp (dist cl rcl));
      x_p_cy_conpoly n q p
    in
    Debug.dprintf4 debug_todo "@[analyse_combi %a: %a@]@\n"
      print_combi1 combi print_conpoly p;
    p


  let analyse =
    fun t age con exp ->
      match exp with
      | ExpCombi (id,th,combi) ->
        let p = analyse_combi t age combi in
        Debug.dprintf5 debug_todo "ExpCombi:%i @[th=%a; p=%a@]@\n"
          id Th.print th print_conpoly p;
        assert (Th.equal th p.exp);
        Conflict.return con conpoly p
      | ExpCstDiff(q,combi) ->
        let p = analyse_combi t age combi in
        Debug.dprintf4 debug_todo "ExpCstDiff: @[q=%a; p=%a@]@\n"
          Q.pp_print q print_conpoly p;
        let q = cst q in
        let p = {imp = add p.imp q; exp = add p.exp q} in
        assert (is_zero p.exp);
        Conflict.return con conpoly p
      | ExpMult(p,g,f,own) ->
        let q = analyse_combi t age (CombiDom(Q.one,g,CombiNull)) in
        let g' = Polynome.(sub_cst (monome Q.one g) p) in
        let q' = x_p_cy_conpoly q Q.minus_one (mk_conpoly g') in
        assert (is_zero q'.exp);
        let fg = monome Q.one own in
        let pf = monome (Q.neg p) f in
        let fg_pf = add fg pf in
        match Polynome.is_cst q'.imp with
        | Some c ->
          let r = {imp = add fg_pf (monome c f);
                   exp = fg_pf} in
          Conflict.return con conpoly r
        | None ->
          (** add a cut: g = p, since g - p - q' = 0; q' = 0  *)
          Conflict.ComputeConflict.unknown_con t conpoly q';
          Conflict.return con conpoly (mk_conpoly fg_pf)

  let expdom t age dom' cl con exp =
      match exp with
      | ExpCombi (_,th,_) ->
        let p = th in
        if ComputeConflict.before_first_dec t age
        then Conflict.return con conpoly (mk_conpoly p)
        else Conflict.return con conpoly {imp = Polynome.zero; exp = p}
      | ExpCstDiff(q,combi) ->
        raise Impossible (** only used for last explanation *)
      | ExpMult (p,g,f,own) ->
        assert false (* todo *)


  let key = expequality

end

module EExpEquality = Conflict.RegisterExp(ExpEquality)

module ExpMulSubst = struct
  open Conflict
  (* open IterExp *)
  open ComputeConflict

  type t = expmulsubst

  let print fmt v =
    Format.fprintf fmt "Subst(%a)" Mul.print v
(*
  let iterexp t age = function
    | Mul(f,g) as v ->
      need_sem t age mul_sem v;
      need_cl_repr t age f;
      need_cl_repr t age g
*)

  let analyse :
    type a. Conflict.ComputeConflict.t ->
    Explanation.age -> a Explanation.con -> t -> a Conflict.rescon =
    fun t age con exp ->
      let get_con () t = function
        | GRequested ()   -> ()
        | GOther (con,c) -> ComputeConflict.unknown_con t con c in
      let get_cons = fold_rescon_list t get_con confact in
      let return () : a Conflict.rescon =
        match Explanation.Con.Eq.eq_type confact con with
        | None -> GOther (confact,())
        | Some Types.Eq -> GRequested () in
      begin match exp with
      | Mul(f,g) as v ->
        GenEquality.equality t age f (get_repr_at t age f);
        GenEquality.equality t age g (get_repr_at t age g);
      end;
      return ()

  let expdom _ _ _ _ _ _ = raise TODO

  let key = expmulsubst

end

module EExpMulsubst = Conflict.RegisterExp(ExpMulSubst)


(** API *)

let index x = Cl.index sem x real

let cst c = index (cst c)

let as_cl cl = monome Q.one cl

let add cl1 cl2 =
  index (add (as_cl cl1) (as_cl cl2))

let sub cl1 cl2 =
  index (sub (as_cl cl1) (as_cl cl2))

let mult cl1 cl2 =
  Cl.index mul_sem (make_mul cl1 cl2) real

let mult_cst cst cl =
  index (mult_cst cst (as_cl cl))
