(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

type bound = Strict | Large


type interval = { lb : bound; lv: Q.t; rv: Q.t; rb: bound }

let print_q fmt q =
  match Q.classify q with
  | Q.ZERO  -> Format.pp_print_char   fmt '0'
  | Q.INF   -> Format.pp_print_string fmt "+∞"
  | Q.MINF  -> Format.pp_print_string fmt  "-∞"
  | Q.UNDEF -> Format.pp_print_string fmt "!undef!"
  | Q.NZERO -> Q.pp_print fmt q

let print_inter fmt t =
  let print_bound_left fmt = function
    | Large  -> Format.fprintf fmt "["
    | Strict -> Format.fprintf fmt "]" in
  let print_bound_right fmt = function
    | Large  -> Format.fprintf fmt "]"
    | Strict -> Format.fprintf fmt "[" in
  Format.fprintf fmt "%a%a;%a%a"
    print_bound_left t.lb
    print_q t.lv
    print_q t.rv
    print_bound_right t.rb

let equal_inter e1 e2 =
  e1.lb == e2.lb && e2.rb == e2.rb && Q.equal e1.lv e2.lv && Q.equal e1.rv e2.rv

type t = interval list
(** all t returned to the outside should verify this invariant *)

let print fmt l =
  Pp.print_list (Pp.constant_string "∪") print_inter fmt l

let invariant l =
  let rec aux minb minv l =
    match minb, l with
    | Large, [] when Q.equal minv Q.inf -> false
    | _    , []                         -> true
    | _     , {lb=Large ; lv}::_ when Q.compare lv minv <= 0 -> false
    | Large , {lb=Strict; lv}::_ when Q.compare lv minv <= 0 -> false
    | Strict, {lb=Strict; lv}::_ when Q.compare lv minv <  0 -> false
    | _, {lb=Strict; lv; rv}::_            when Q.compare rv lv <= 0 -> false
    | _, {lb=Large ; lv; rv; rb=Strict}::_ when Q.compare rv lv <= 0 -> false
    | _, {lb=Large ; lv; rv; rb=Large }::_ when Q.compare rv lv <  0 -> false
    | _, {rv;rb}::l -> aux rb rv l
  in
  aux Strict Q.minus_inf l


let empty = []
let is_empty t = t == []
let reals = [{lb=Strict; lv=Q.minus_inf; rb=Strict; rv=Q.inf}]
let is_reals = function
  | [{lb=Strict; lv; rb=Strict; rv}] when Q.equal lv Q.minus_inf &&
                                          Q.equal rv Q.inf     -> true
  | _ -> false

let is_singleton = function
  | [{lv;rv}] when Q.equal lv rv -> Some lv
  | _ -> None

let min_max e1 l1 t1 e2 l2 t2 =
  let c = Q.compare e1.lv e2.lv in
  match e1.lb, e2.lb with
  | Strict, Large
    when c =  0 -> e2,l2,t2,e1,l1,t1
  | _
    when c <= 0 -> e1,l1,t1,e2,l2,t2
  | _           -> e2,l2,t2,e1,l1,t1

let rec is_distinct t1 t2 =
  match t1,t2 with
  | [], _ | _, [] -> true
  | e1::l1, e2::l2 ->
  (** order by the minimum *)
  let emin,lmin,_,emax,_,tmax = min_max e1 l1 t1 e2 l2 t2 in
  (** look for inclusion *)
  let c = Q.compare emin.rv emax.lv in
  match emin.rb, emax.lb with
  | Strict, Strict | Strict, Large | Large, Strict
    when c <= 0 -> is_distinct lmin tmax
  | Large, Large
    when c <  0 -> is_distinct lmin tmax
  | _ -> false


let rec union t1 t2 =
  match t1,t2 with
  | [], l | l, [] -> l
  | e1::l1, e2::l2 ->
  (** order by the minimum *)
  let emin,lmin,_,emax,lmax,tmax = min_max e1 l1 t1 e2 l2 t2 in
  (** look for an intersection *)
  let c = Q.compare emin.rv emax.lv in
  match emin.rb, emax.lb with
  | Strict, Strict
    when c <= 0 -> emin::(union lmin tmax)
  | Large,Large | Strict, Large | Large, Strict
    when c <  0 -> emin::(union lmin tmax)
  | _ ->
    (** look for inclusion *)
    let c = Q.compare emax.rv emin.rv in
    match emin.lb, emax.lb with
    (** max included in min *)
    | Strict, Strict | Large, Large | Large, Strict
      when c <= 0 -> emin::(union lmin lmax)
    | Strict, Large
      when c < 0 -> emin::(union lmin lmax)
    (** merge the two *)
    | _ ->
      let e = {lv = emin.lv; lb = emin.lb; rv = emax.rv; rb = emax.rb} in
      union lmin (e::lmax)


let rec inter' t1 t2 =
  match t1,t2 with
  | [], _ | _, [] -> []
  | e1::l1, e2::l2 ->
  (** order by the minimum *)
  let emin,lmin,tmin,emax,lmax,tmax = min_max e1 l1 t1 e2 l2 t2 in
  (** look for an intersection *)
  let c = Q.compare emin.rv emax.lv in
  match emin.rb, emax.lb with
  | Strict, Strict
    when c <= 0 -> inter' lmin tmax
  | Large,Large | Strict, Large | Large, Strict
    when c <  0 -> inter' lmin tmax
  | _ ->
    (** look for inclusion *)
    let c = Q.compare emax.rv emin.rv in
    match emin.lb, emax.lb with
    (** max included in min *)
    | Strict, Strict | Large, Large | Large, Strict
      when c <= 0 -> emax::(inter' tmin lmax)
    | Strict, Large
      when c < 0 -> emax::(inter' tmin lmax)
    (** merge the two *)
    | _ ->
      let e = {lv = emax.lv; lb = emax.lb; rv = emin.rv; rb = emin.rb} in
      e::(inter' lmin tmax)

(** special case if the two are equals, return the second *)
let rec inter t1 t2 =
  match t1, t2 with
  | _ when t1 == t2 -> t2
  | [], _ | _, [] -> []
  | e1::l1, e2::l2 when equal_inter e1 e2 ->
    let l = inter l1 l2 in
    if l == l2 then t2 else e2::l2
  | _ -> inter' t1 t2

let singleton q =
  let t = [{lb=Large; lv = q; rv = q; rb= Large}] in
  assert (invariant t);
  t

let gt q =
  let t = [{lb=Strict; lv = q; rv = Q.inf; rb= Strict}] in
  assert (invariant t);
  t

let ge q =
  let t = [{lb=Large ; lv = q; rv = Q.inf; rb= Strict}] in
  assert (invariant t);
  t

let lt q =
  let t = [{lb=Strict; lv = Q.minus_inf; rv = q; rb= Strict}] in
  assert (invariant t);
  t

let le q =
  let t = [{lb=Strict; lv = Q.minus_inf; rv = q; rb= Large }] in
  assert (invariant t);
  t

let rec add_cst q = function
  | []   -> []
  | a::l -> {a with lv = Q.add a.lv q; rv = Q.add a.rv q}::(add_cst q l)

let add_cst q t =
  if Q.equal Q.zero q then t else add_cst q t

let rec mult_pos q = function
  | []   -> []
  | a::l -> {a with lv = Q.mul a.lv q; rv = Q.mul a.rv q}::(mult_pos q l)

let rec mult_neg acc q = function
  | []   -> acc
  | a::l -> mult_neg ({lb = a.rb; rb = a.lb;
                       lv = Q.mul a.rv q;
                       rv = Q.mul a.lv q}::acc) q l

let mult_cst q t =
  assert (Q.is_real q);
  let c = Q.sign q in
  if c = 0 then singleton Q.zero
  else if c > 0 then mult_pos q t
  else               mult_neg [] q t


let mult_bound b1 b2 =
  match b1, b2 with
  | Large , Large  -> Large
  | _              -> Strict

(** t is smaller than l but perhaps a merge is needed *)
let cons t l =
  match t.rb, l with
  | _,[] -> [t]
  | Strict, ({lb=Strict} as e)::_ when Q.compare t.rv e.lv <= 0 ->
    t::l
  | _, e::_                       when Q.compare t.rv e.lv <  0 ->
    t::l
  | _, e::l ->
    assert (Q.compare t.lv e.lv < 0);
    assert (Q.compare t.rv e.rv < 0);
    {lb=t.lb; lv = t.rv; rv = e.rv; rb = e.rb}::l

let rec add_interval t = function
  | [] -> []
  | e::l ->
    let e = {lb=mult_bound t.lb e.lb;
             lv=Q.add t.lv e.lv;
             rv=Q.add t.rv e.rv;
             rb=mult_bound t.rb e.rb} in
    cons e (add_interval t l)

let add t1 t2 =
  let res = match is_singleton t1, t1, is_singleton t2, t2 with
    | None,_, None,_ ->
      List.fold_left (fun acc t -> union acc (add_interval t t2)) [] t1
    | Some q, _, None, t
    | None, t, Some q, _ when Q.equal Q.zero q -> t
    | Some q, _, None, t
    | None, t, Some q, _ ->
      add_cst q t
    | Some q1,_, Some q2,_ -> singleton (Q.add q1 q2)
  in
  assert ( invariant res );
  res

(** TODO better heuristic *)
let choose ?exn = function
  | [] -> raise (Opt.get_def Not_found exn)
  | {lb=Large;lv}::_ -> lv
  | {rb=Large;rv}::_ -> rv
  | {lv;rv}::_ when Q.equal Q.minus_inf lv ->
    Q.make (Z.sub Z.one (Q.to_bigint rv)) Z.one
  | {lv;rv}::_ when Q.equal Q.inf rv ->
    Q.make (Z.add Z.one (Q.to_bigint lv)) Z.one
  | {lv;rv}::_ -> Q.add rv (Q.div_2exp (Q.sub lv rv) 1)


let mult_cst q t =
  let res = mult_cst q t in
  assert ( invariant res );
  res

let union t1 t2 =
  let res = union t1 t2 in
  assert ( invariant res );
  res

let inter t1 t2 =
  let res = inter t1 t2 in
  assert ( invariant res );
  res

let unit_test () =
  assert ( invariant empty );
  assert ( is_empty empty );
  assert ( invariant reals );
  assert ( is_reals reals );
  let zero = singleton Q.zero in
  let one  = singleton Q.one  in
  let two  = singleton (Q.of_int 2)  in
  let three  = singleton (Q.of_int 3)  in
  assert (not (is_distinct zero zero));
  assert (is_distinct zero one);
  let one_three = inter (ge Q.one) (lt (Q.of_int 3)) in
  assert (not (is_distinct two one_three));
  assert (is_distinct three one_three);
  assert (not (is_distinct one one_three));
  ()

let () = unit_test ()
