(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Popop
open Types


module T = struct
  type t = { cst : Q.t; poly : Q.t Cl.M.t}

  let equal n m = Q.equal n.cst m.cst && Cl.M.equal Q.equal n.poly m.poly

  let hash n = (** number au pif *)
    Cl.M.fold (fun k v acc ->
        Cl.hash k * 101 + Hashtbl.hash v * 107 + acc * 253)
      n.poly (Hashtbl.hash n.cst * 27)

  let compare n m =
    let c = Q.compare n.cst m.cst in
    if c <> 0 then c
    else Cl.M.compare Q.compare n.poly m.poly

  let print fmt v =
    let print_not_1 first fmt q =
      if not first && Q.compare q Q.zero >= 0
      then Format.pp_print_string fmt "+";
      if Q.equal q Q.zero then  Format.pp_print_string fmt "!0!"
      else if Q.equal Q.minus_one q then Format.pp_print_string fmt "-"
      else if not (Q.equal Q.one q) then Q.pp_print fmt q
    in
    let print_not_0 first fmt q =
      if first
      then Q.pp_print fmt q
      else
      if not (Q.equal Q.zero q) then begin
        if Q.compare q Q.zero > 0 then Format.pp_print_string fmt "+";
        Q.pp_print fmt q
      end
    in
    let print_mono k v (fmt,first) =
      Format.fprintf fmt "@[%a%a@]@," (print_not_1 first) v Cl.print k;
      (fmt,false)
    in
    Format.fprintf fmt "@[";
    let _,first = Cl.M.fold print_mono v.poly (fmt,true) in
    Format.fprintf fmt "%a@]" (print_not_0 first) v.cst

end

open T
include Stdlib.MkDatatype(T)

(** different invariant *)

let invariant p =
  Cl.M.for_all (fun _ q -> not (Q.equal q Q.zero)) p.poly

(** constructor *)
let cst q = {cst = q; poly = Cl.M.empty}
let zero = cst Q.zero
let is_cst p = if Cl.M.is_empty p.poly then Some p.cst else None
let is_zero p = Q.equal p.cst Q.zero && Cl.M.is_empty p.poly

type extract = Zero | Cst of Q.t | Var of Q.t * Cl.t * t
let extract p =
  if Cl.M.is_empty p.poly then
    if Q.equal p.cst Q.zero then Zero
    else Cst p.cst
  else
    let x,q = Shuffle.chooseb Cl.M.choose Cl.M.choose_rnd p.poly in
    let p' = {p with poly = Cl.M.remove x p.poly} in
    Var(q,x,p')

type kind = ZERO | CST | VAR
let classify p =
  if Cl.M.is_empty p.poly then
    if Q.equal p.cst Q.zero then ZERO
    else CST
  else
    VAR


let monome c x =
  if Q.equal Q.zero c then cst Q.zero
  else {cst = Q.zero; poly = Cl.M.singleton x c}

let is_one_cl p = (** cst = 0 and one empty monome *)
  if Q.equal Q.zero p.cst && Cl.M.is_num_elt 1 p.poly then
    let cl,k = Cl.M.choose p.poly in
    if Q.equal Q.one k then Some cl
    else None
  else None

let sub_cst p q = {p with cst = Q.sub p.cst q}

let mult_cst c p1 =
  if Q.equal Q.one c then p1
  else
  let poly_mult c m = Cl.M.map (fun c1 -> Q.mul c c1) m in
  if Q.equal Q.zero c then cst Q.zero
  else {cst = Q.mul c p1.cst; poly = poly_mult c p1.poly}


let none_zero c = if Q.equal Q.zero c then None else Some c

(** Warning Cl.M.union can be used only for defining an operation [op]
    that verifies [op 0 p = p] and [op p 0 = p] *)
let add p1 p2 =
  let poly_add m1 m2 =
    Cl.M.union (fun _ c1 c2 -> none_zero (Q.add c1 c2)) m1 m2
  in
  {cst = Q.add p1.cst p2.cst; poly = poly_add p1.poly p2.poly}

let sub p1 p2 =
  let poly_sub m1 m2 =
    Cl.M.union_merge (fun _ c1 c2 ->
      match c1 with
      | None -> Some (Q.neg c2)
      | Some c1 -> none_zero (Q.sub c1 c2))
      m1 m2 in
  {cst = Q.sub p1.cst p2.cst; poly = poly_sub p1.poly p2.poly}

let x_p_cy p1 c p2 =
  assert (not (Q.equal c Q.zero));
  let f a b = Q.add a (Q.mul c b) in
  let poly m1 m2 =
    Cl.M.union_merge (fun _ c1 c2 ->
      match c1 with
      | None -> Some (Q.mul c c2)
      | Some c1 -> none_zero (f c1 c2))
      m1 m2 in
  {cst = f p1.cst p2.cst; poly = poly p1.poly p2.poly}


let subst_cl p x y =
  let poly,qo = Cl.M.find_remove x p.poly in
  match qo with
  | None -> p, Q.zero
  | Some q ->
    let poly = Cl.M.change (function
        | None -> qo
        | Some q' -> none_zero (Q.add q q')
      ) y poly in
    {p with poly}, q

let subst p x s =
  let poly,q = Cl.M.find_remove x p.poly in
  match q with
  | None -> p, Q.zero
  | Some q -> x_p_cy {p with poly} q s, q

let fold f acc p = Cl.M.fold_left f acc p.poly
let iter f p = Cl.M.iter f p.poly
