(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)
open Popop
open Stdlib
open Types
open Solver

let lazy_propagation = false



let debug = Debug.register_info_flag
  ~desc:"for bit-vectors theory"
  "bv"

(*
  Some sections of the code are tagged, like the following.
  This is used for listing in the documentation
*)

(* Bit *)
type binop = | Bxor | Bor | Band | Bsl | Bsr
type bvexp =
| Binop of binop * Cl.t * Cl.t
| Bnot of Cl.t
| Cst of Z.t
(* end *)

let hash_bop a b = function
| Bxor -> 3 * a + 5 * b
| Bor -> 7 * a + 9 * b
| Band -> 11 * a + 13 * b
| Bsl ->  19 * a + 31 * b
| Bsr -> 23 * a + 33 * b

let sem : bvexp sem = Sem.create_key "bv_sem"

(* BVdomain *)
type bvdom =   {aun : Z.t; azer : Z.t}
(* end *)
let dom : bvdom dom =   Dom.create_key "bvdom"

let domlen x = max (Z.numbits x.aun) (Z.numbits x.azer)
let bvmeet x y = Z.({aun = logor x.aun y.aun ; azer = logand x.azer  y.azer})
let bneg x = Z.({aun = lognot x.azer ; azer = lognot x.aun})
let bveqdom x y = Z.equal x.aun y.aun && Z.equal x.azer y.azer
let bvconflict x = not (Z.equal Z.(logand (lognot x.azer)  x.aun) Z.zero)
let normalshift shift_op x n =
  Z.({aun = shift_op x.aun n; azer = shift_op x.azer n})

let shifted_ones shift_op x n =
if (shift_op Z.one 1 == (Z.of_int 2)) then (* it's << *)
  Z.lognot (shift_op  Z.minus_one n)
else 					(* it's >>, so use << *)
  let len = domlen x in
  Z.shift_left Z.minus_one (len - n)

let unknownshift shift_op x n =
  let ones = shifted_ones shift_op x n in
  Z.({aun = shift_op x.aun n;
      azer = logor (shift_op x.azer n) ones })

let bvfixed x = if Z.equal x.aun x.azer then Some x.aun else None

module D = struct
  type t = bvdom

  type expmerge = (* all Copied *)
  | Merge of Explanation.pexp * Cl.t * Cl.t
  | DomMerge of Explanation.pexp * Cl.t

  let expmerge : expmerge Explanation.exp =(* nearly all Copied *)
    Explanation.Exp.create_key "Bit-vector.merge"

  let set_dom_bv d pexp cl newdom =
    match Delayed.get_dom d dom cl with
    | Some b' when bveqdom b' newdom -> ()
    | _ -> Delayed.set_dom d pexp dom cl newdom

  let get_dom_bv d dom cl =
    match Delayed.get_dom d dom cl with
    | Some b -> b
    | None -> {aun = Z.zero; azer = Z.minus_one}

  let merge d pexp (b1, cl1) (b2, cl2)  _ =
    match b1,b2 with
    | Some v1, Some v2 when bveqdom v1 v2 -> ()
    | Some v1, Some v2 ->
      let newdom = bvmeet v1 v2 in
      if bvconflict newdom then
	let pexp = Delayed.mk_pexp d expmerge (Merge(pexp, cl1, cl2)) in
	Delayed.contradiction d pexp
      else begin
	set_dom_bv d pexp cl1 newdom;
	set_dom_bv d pexp cl2 newdom;
      end
    | None, Some b -> Delayed.set_dom_premerge d dom cl1 b
    | Some b, None -> Delayed.set_dom_premerge d dom cl2 b
    | None, None -> failwith "Should not be possible"

  let key = dom

  let print_cnst fmt cbv =
    let to_print = "["^ (Z.format "%b" cbv) ^"]" in
    Format.pp_print_string fmt to_print

  let print fmt interval =
    let len = 1 + domlen interval in
    (** Here we print most significant bit first *)
    let to_print =
      "["^ String.init len
        (fun i ->
          match Z.testbit interval.aun  (len-1-i),
            Z.testbit interval.azer (len-1-i) with
            | true, true -> '1'
            | false, false -> '0'
            | false, true -> '?'
            | true, false -> 'X'
        ) ^ "]" in
    Format.pp_print_string fmt to_print

  let merged b1 b2 =
    match  b1,b2 with
    | Some v1, Some v2 -> bveqdom v1 v2
    | _, _ -> false

end

module DE = RegisterDom(D)

(* Useful utils: removable *)
let simpleprint var env str=
  let xd = D.get_dom_bv  env dom var in
  Format.print_string ("\n"^ str);
  D.print Format.std_formatter xd;
  Format.print_string "\n"
let simpledomprint xd str=
  Format.print_string ("\n"^ str);
  D.print Format.std_formatter xd;
  Format.print_string "\n"

module T = struct
  type t = bvexp
  let equal e1 e2 =
    match e1,e2 with
    | Binop (bop, a, b), Binop (bop', aa, bb) ->
      (bop == bop') && Cl.equal a aa && Cl.equal b bb
    | Cst i, Cst j -> Z.equal i j
    | Bnot a , Bnot aa -> Cl.equal a aa
    | _,_ -> false

  let hash = function
    | Binop (bop, a, b) -> hash_bop (Cl.hash a) (Cl.hash b) bop
    | Bnot a -> 17 * Cl.hash a
    | Cst i -> Z.hash i

  let compare n m  =
    match n, m with
    | Binop (bop, a, b), Binop (bop', aa, bb) ->
      let c = compare bop bop' in
      if c <> 0 then c
      else let c = Cl.compare a aa in
	   if c <> 0 then c
	   else Cl.compare b bb
    | Binop _ , _ -> 1
    | _ , Binop _ -> 1
    | Bnot a, Bnot aa -> Cl.compare a aa
    | Bnot _, _ -> 1
    | _, Bnot _ -> 1
    | Cst i, Cst j -> Z.compare i j

  let op_to_string = function
   | Bxor -> "⊕"
   | Bor -> "|"
   | Band -> "&"
   | Bsl -> "<<"
   | Bsr -> ">>"

  let print fmt (e : t)=
    match e with
    | Binop (bop, a, b) ->
      Format.fprintf fmt ("(%a) %s (%a)") Cl.print a (op_to_string bop) Cl.print b
    | Bnot a ->
      Format.fprintf fmt "~(%a)" Cl.print a
    | Cst i -> D.print_cnst fmt  i
end

module Th = struct
  include MkDatatype(T)

  let key = sem

end

module ThE = RegisterSem(Th)

let and_dirp x y =  Z.({aun = x.aun land y.aun ; azer = x.azer land y.azer})
let and_invp x z = Z.({aun = z.aun ; azer = logor z.azer  (lognot x.aun)})
let or_dirp  x y =  Z.({aun = x.aun lor y.aun ; azer = x.azer lor y.azer})
let or_invp x z = Z.({aun = z.aun land (lognot x.azer) ; azer = z.azer})
let xor_p x y =
  Z.({aun  = (((lognot x.azer) land y.aun) lor (x.aun land (lognot y.azer))) ;
      azer = (x.azer lor y.azer) land lognot (x.aun land y.aun)})

let shift_p sh_op inv_op x n z =
  let nx = bvmeet x (unknownshift inv_op z n) in
  let nz = bvmeet z (normalshift sh_op x n) in
  if List.exists bvconflict [nx;nz] then None
  else Some (nx, nz)

let dom_of_cnst i = {aun = i; azer = i}

let or_prop x y z = (* x | y = z *)
  let nz = bvmeet z (or_dirp x y) in
  let nx = bvmeet x (or_invp y z) in
  let ny = bvmeet y (or_invp x z) in
  if List.exists bvconflict [nx;ny;nz] then None
  else Some (nx, ny, nz)

let and_prop x y z = (* x & y = z *)
  let nz = bvmeet z (and_dirp x y) in
  let nx = bvmeet x (and_invp y z) in
  let ny = bvmeet y (and_invp x z) in
  if List.exists bvconflict [nx;ny;nz] then None
  else Some (nx, ny, nz)

let xor_prop x y z =
  let nx = bvmeet x (xor_p y z) in
  let ny = bvmeet y (xor_p x z) in
  let nz = bvmeet z (xor_p x y) in
  if List.exists bvconflict [nx;ny;nz] then None
  else Some (nx, ny, nz)

type side =
| Left
| Right
| Up
type expprop =
| ExpDirBV of ThE.t
| ExpInvBV of ThE.t * side
| ExpConflictBV of ThE.t

let expprop : expprop Explanation.exp =
  Explanation.Exp.create_key "Bv.prop"

module ExpProp = struct  		(* TODO *)
  open D
  type t = expprop
  let key = expprop

  let analyse _ _ _ _  =  failwith "analyse exp"

  let print _ = failwith "print expdom"

  let expdom _ _ _ _ _ _ = failwith "expdom"

end

module EP = Conflict.RegisterExp(ExpProp)

let ty_ctr = Ty.Constr.create "Bit-vector"
let ty = Ty.ctr ty_ctr

let index a = Cl.index sem a ty

module DaemonPropa = struct
  type info = ThE.t

  module Data = struct
    type t = info
    let print  fmt clsem =
      Format.fprintf fmt "Propagating inside %a" ThE.print clsem
  end

  let key = Demon.Fast.create "BV.DaemonPropa"
  let immediate = true
  let throttle = 100

  let dom_set_or_merge ds exp c domc =
    match bvfixed domc with
    |None -> D.set_dom_bv ds exp c domc
    |Some i ->
      let cst = index (Cst i) in
      Solver.Delayed.register ds cst;
      Solver.Delayed.flush ds;
      (* let pexp = Explanation.pexpfact in *)
      (* D.set_dom_bv ds pexp cst {aun = i; azer = i}; *)
      Delayed.merge ds exp c cst

  let generic_shift d sh_op inv_op clsem c off =
    let offset = D.get_dom_bv d dom off in
    match bvfixed offset with
    | None -> ()
    | Some nn ->
      let n = Z.to_int nn in
      let cl = ThE.cl clsem in
      let dc = D.get_dom_bv d dom c in
      let dcl = D.get_dom_bv d dom cl in
      match shift_p sh_op inv_op dc n dcl with
      | None ->
	let pexp = Delayed.mk_pexp d expprop (ExpConflictBV clsem) in
	Delayed.contradiction d pexp
      | Some (ndc, ndcl) ->
	let pexp = Delayed.mk_pexp d expprop (ExpDirBV clsem) in
	let pexp1 = Delayed.mk_pexp d expprop  (ExpInvBV (clsem, Left)) in
	(* I also want an explanation that says
	   "it's because the offset became fixed" *)
	dom_set_or_merge d pexp1  c ndc;
	dom_set_or_merge d pexp cl ndcl

  let generic_prop d prop_fun clsem c1 c2 =
    let cl = ThE.cl clsem in
    let dcl = D.get_dom_bv d dom cl in
    let dc1 = D.get_dom_bv d dom c1 in
    let dc2 = D.get_dom_bv d dom c2 in
    match prop_fun dc1 dc2 dcl with
    | None ->
      let pexp = Delayed.mk_pexp d expprop (ExpConflictBV clsem) in
      Delayed.contradiction d pexp
    | Some (ndc1, ndc2, ndcl) ->
      let pexp = Delayed.mk_pexp d expprop (ExpDirBV clsem) in
      let pexp1 = Delayed.mk_pexp d expprop  (ExpInvBV (clsem, Left)) in
      let pexp2 = Delayed.mk_pexp d expprop (ExpInvBV (clsem, Right)) in
      dom_set_or_merge d pexp  cl ndcl;
      dom_set_or_merge d pexp1 c1 ndc1;
      dom_set_or_merge d pexp2 c2 ndc2

  let wakeup d = function
    | Events.Fired.EventDom(_, dom', clsem) ->
      assert (Dom.equal dom dom');
      begin match ThE.sem clsem with
      | Bnot c ->
	let cl = ThE.cl clsem in
	let dcl = D.get_dom_bv d dom cl in
	let dc = D.get_dom_bv d dom c in
	let ndcl = bvmeet dcl (bneg dc) in
	let ndc = bvmeet dc (bneg dcl) in
	if List.exists bvconflict [ndcl; ndc] then
	  let pexp = Delayed.mk_pexp d expprop (ExpConflictBV clsem) in
	  Delayed.contradiction d pexp
	else
	  let pexp = Delayed.mk_pexp d expprop (ExpDirBV clsem) in
	  let pexp1 = Delayed.mk_pexp d expprop  (ExpInvBV (clsem, Up)) in
	  D.set_dom_bv d pexp  cl ndcl;
	  D.set_dom_bv d pexp1 c ndc
      | Binop (Bor, a, b) -> generic_prop d or_prop clsem a b
      | Binop (Band, a, b) -> generic_prop d and_prop clsem a b
      | Binop (Bxor, a, b) -> generic_prop d xor_prop clsem a b
      | Binop (Bsl, c, off) -> generic_shift d (Z.shift_left) (Z.shift_right) clsem c off
      | Binop (Bsr, c, off) -> generic_shift d (Z.shift_right) (Z.shift_left) clsem c off
      | _ -> failwith "should be impossible"
      end
    | _ -> raise UnwaitedEvent

  let init d clsem =
    let v = ThE.sem clsem in
    let ctxt = ThE.cl clsem in
    match v with
    | Binop (_, a, b) ->
      let events = [Demon.Create.EventDom (a,dom,clsem);
                    Demon.Create.EventDom (b,dom,clsem);
		    Demon.Create.EventDom (ctxt,dom,clsem)] in
      Demon.Fast.attach d key events;
      wakeup d (Events.Fired.EventDom(ctxt, dom, clsem))
    | Bnot c  ->
      let events = [Demon.Create.EventDom (c,dom,clsem);
		    Demon.Create.EventDom (ctxt,dom,clsem)] in
      Demon.Fast.attach d key events;
      wakeup d (Events.Fired.EventDom(ctxt, dom, clsem))
    | Cst i ->
      let pexp = Explanation.pexpfact in
      D.set_dom_bv d pexp ctxt {aun = i; azer = i}
end

module RDaemonPropa = Demon.Fast.Register(DaemonPropa)

module DaemonInit = struct

  let key = Demon.Fast.create "BV.DaemonInit"
  type d = unit
  module Data = DUnit
  let immediate = true
  let throttle = 100

  let wakeup d  =  function
    | Events.Fired.EventRegSem(clsem,()) ->
      let clsem = ThE.coerce_clsem clsem in
      begin
	try
	  DaemonPropa.init d clsem
	with Exit -> ()
      end
    | _ -> raise UnwaitedEvent
end

module RDaemonInit = Demon.Fast.Register(DaemonInit)

module ExpMerge = struct 		(* TODO *)
  open D

  type t = expmerge
  let key = expmerge

  let expdom _ _ _ _ _ _ = raise Impossible (* used only for unsat *)
  let analyse _ _ =   failwith "Zak: analyse in Exp"

  let print fmt = function
    | Merge  (pexp,cl1,cl2)   ->
      Format.fprintf fmt "Merge!(%a,%a,%a)"
        Conflict.print_pexp pexp Cl.print cl1 Cl.print cl2
    | DomMerge (pexp,cl) ->
      Format.fprintf fmt "DomMerge!(%a,%a)"
        Conflict.print_pexp pexp Cl.print cl
end

module EM = Conflict.RegisterExp(ExpMerge)

let chobv = Explanation.Cho.create_key "Bv.cho"

module ChoBv = struct			(* TODO ? *)
  open Conflict

  module Key = Cl
  module Data = D
  let key = chobv
  let analyse (type a) t (con: a Explanation.con) cl b = failwith "todo"
  let make_decision env dec cl b = () (* failwith "todo Decision" *)
  let choose_decision env cl =
    let dcl= D.get_dom_bv env dom cl in
    match Z.(dcl.aun lxor dcl.azer) with
    |i when Z.equal i Z.zero -> DecNo
    |n -> DecTodo Z.({aun = n lor dcl.aun; azer = dcl.azer})
end

module EChoBv = Conflict.RegisterCho(ChoBv)

let make_dec cl = Explanation.GCho (chobv,cl)
let () = Variable.register_sort  ~dec:make_dec  ty

let th_register ?(with_other_theories=false) env =
  RDaemonPropa.init env;
  RDaemonInit.init env;
  Demon.Fast.attach env
    DaemonInit.key [Demon.Create.EventRegSem(sem,())];
  ()

(* Utilities *)

let cnst c = index  (Cst (Z.of_int c))
let band a b = index (Binop (Band ,a, b))
let bor a b = index (Binop (Bor, a, b))
let bxor a b = index (Binop (Bxor, a, b))
let bsr a b = index (Binop (Bsr, a, b))
let bsl a b = index (Binop (Bsl, a, b))
let bnot a = index (Bnot a)
