(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Popop
open Types

let debug = Debug.register_info_flag
  ~desc:"for the specialized demons"
  "Demon.all"


module Create = struct
    type 'b event =
    | EventDom      : Cl.t * 'a dom  * 'b -> 'b event
    | EventRegCl  : Cl.t           * 'b -> 'b event
    | EventChange   : Cl.t           * 'b -> 'b event
    | EventRegSem :        'a sem  * 'b -> 'b event


    let print fmt = function
      | EventDom      (cl, dom, _) ->
        Format.fprintf fmt "dom:%a of %a" Dom.print dom Cl.print cl
      | EventRegCl  (cl, _)    ->
        Format.fprintf fmt "registration of %a" Cl.print cl
      | EventChange   (cl, _)    ->
        Format.fprintf fmt "changecl of %a" Cl.print cl
      | EventRegSem (sem, _)    ->
        Format.fprintf fmt "regsem for %a" Sem.print sem


    type 'b t = 'b event list
end

type 'k alive =
| AliveReattached
| AliveStopped
| AliveRedirected of 'k

module Key = struct

  type ('d,'k,'i) daemon_state =
    | Alive of 'd Solver.Events.Fired.t * 'i
    | Dead
    | Redirected of 'k

  let print_daemon_state fmt = function
    | Alive _ -> Format.fprintf fmt "alive"
    | Dead -> Format.fprintf fmt "dead"
    | Redirected _ -> Format.fprintf fmt "redirected"

  module type DemTable = sig
    module Key: Stdlib.Datatype
    type data
    type info val default: info
    val state : (data, Key.t, info) daemon_state Key.M.t
  end

  type ('k,'d,'i) demtable =
    (module DemTable with type Key.t = 'k and type data = 'd
                                          and type info = 'i)

  type ('k,'d,'i) t = {
    dk_id : ('k * 'd, 'k) Types.dem;
    dk_data : ('k,'d,'i) demtable Types.env;
  }

  let create name = {
    dk_id   = Dem.create_key name;
    dk_data = Env.create_key name;
  }

  module type S = sig
    module Key: Stdlib.Datatype

    module Data: Stdlib.Printable

    type info
    val default: info

    val key: (Key.t, Data.t, info) t

    val immediate: bool
    val wakeup:
      Solver.Delayed.t -> Key.t -> Data.t Solver.Events.Fired.t ->
      info -> Key.t alive
    (** the Events.t in wakeup is a subset of the one given in watch *)
  end

  (** mark it attached if it is not already the case *)
  let mark_dem :
  type k d i. Solver.Delayed.t -> (k,d,i) t -> k -> unit =
    fun d dem k ->
      try
        let module DemTable = (val (Solver.Delayed.get_env d dem.dk_data)) in
        let module DemTable' = struct
          include DemTable
          let state = DemTable.Key.M.change (function
              | None -> Some (Alive([],DemTable.default))
              | Some Dead -> raise AlreadyDead
              | Some (Redirected _) -> raise AlreadyRedirected
              | Some (Alive _) -> raise Exit)
              k DemTable.state
        end in
        Solver.Delayed.set_env d dem.dk_data (module DemTable')
      with Exit -> ()

  module Register(D:S) = struct

    let rec run d k =
      let module DemTable = (val (Solver.Delayed.get_env d D.key.dk_data)) in
      match DemTable.Key.M.find k (DemTable.state) with
      | Dead ->
        Debug.dprintf4 debug "[Demon] @[Daemon %a for %a is dead@]@\n"
          Dem.print D.key.dk_id DemTable.Key.print k;
        None
      | Redirected k' ->
        Debug.dprintf6 debug
        "[Demon] @[Daemon %a for %a is redirected to %a@]@\n"
        Dem.print D.key.dk_id DemTable.Key.print
        k DemTable.Key.print k';
        run d k'
      | Alive (events,info) ->
        Debug.dprintf6 debug "[Demon] @[Run daemon %a for %a:@\n@[%a@]@]@\n"
          Dem.print D.key.dk_id DemTable.Key.print k
          (Pp.print_list Pp.newline Solver.Events.Fired.print) events;
        (** event can be added during wakeup *)
        let module DemTable' = struct
          include DemTable
          let state = DemTable.Key.M.add k (Alive([],info)) (DemTable.state)
        end
        in
        Solver.Delayed.set_env d D.key.dk_data (module DemTable');
        (** wakeup *)
        let alive = D.wakeup d k events info in
        (** delayed can be modified *)
        begin match alive with
          | AliveStopped | AliveRedirected _ ->
            let demstate = match alive with
              | AliveStopped -> Dead
              | AliveRedirected k' -> mark_dem d D.key k'; Redirected k'
              | AliveReattached -> assert false  in
            Debug.dprintf4 debug "[Demon] @[Stop daemon %a %a@]@\n"
              Dem.print D.key.dk_id DemTable.Key.print k;
            begin
              let module DemTable =
                (val (Solver.Delayed.get_env d D.key.dk_data)) in
              (** Dead even if event have been added *)
              let state' = DemTable.Key.M.add k demstate (DemTable.state) in
              let module DemTable' = struct
                include DemTable
                let state = state'
              end
              in
              Solver.Delayed.set_env d D.key.dk_data (module DemTable')
            end
          | AliveReattached ->
            Debug.dprintf0 debug "[Demon] @[Reattach daemon@]@\n";
        end;
        None

    let enqueue d event =
      let module DemTable = (val (Solver.Ro.get_env d D.key.dk_data)) in
      let change_state k l =
          Debug.dprintf6 debug
          "[Demon] @[schedule %a for %a with %a@]@\n"
          Dem.print D.key.dk_id D.Key.print k
          Solver.Events.Fired.print event;
        let module DemTable' = struct
          include DemTable
          let state = DemTable.Key.M.add k l DemTable.state
        end in
        Solver.Ro.set_env d D.key.dk_data (module DemTable')
      in
      let rec update_state k data =
        match DemTable.Key.M.find_opt k DemTable.state with
        | None -> assert false (* should have been marked *)
        | Some Dead ->
          Debug.dprintf4 debug
            "[Demon] @[Dem %a is dead for %a@]@\n"
            Dem.print D.key.dk_id Solver.Events.Fired.print event;
          Solver.EnqStopped
        | Some (Redirected k') -> update_state k' data
        | (Some Alive([],info))  ->
          change_state k (Alive([data],info));
          Solver.EnqRun k
        | Some Alive(l,info) ->
          change_state k (Alive(data::l,info));
          Solver.EnqAlready
      in
      let k, event =
        let open Solver.Events.Fired in
        match event with
        | EventDom      (a, b , (k,d))   -> k, EventDom(a, b, d)
        | EventSem      (a, b, c, (k,d)) -> k, EventSem(a, b, c, d)
        | EventReg      (a, (k,d))       -> k, EventReg(a, d)
        | EventRegCl    (a, (k,d))       -> k, EventRegCl(a, d)
        | EventChange   (a, (k,d))       -> k, EventChange(a, d)
        | EventRegSem (a, (k,d))       -> k, EventRegSem(a, d) in
      update_state k event


    let () =
      let print_demtable fmt (d: (D.Key.t,D.Data.t,D.info) demtable) =
        let module DT = (val d) in
        Pp.print_iter2 DT.Key.M.iter Pp.newline Pp.colon
          D.Key.print print_daemon_state fmt DT.state
      in
      Solver.register_env print_demtable D.key.dk_data;
    (** Interface for generic daemon *)
    let module Dem = struct
      type runable = D.Key.t
      let print_runable = D.Key.print
      let run = run

      type event = D.Key.t * D.Data.t
      let print_event fmt (k,d) =
        Format.fprintf fmt "(%a: %a)" D.Key.print k D.Data.print d
      let enqueue = enqueue

      let key = D.key.dk_id
      let immediate = D.immediate
    end in
    let module RDem = Solver.RegisterDem(Dem) in
    ()

    let init d =
      let module DemTable = struct
        module Key = D.Key
        type data = D.Data.t
        type info = D.info let default = D.default
        let state = Key.M.empty
      end in
      Solver.Delayed.set_env d D.key.dk_data (module DemTable);

  end

  let attach :
    type k d i. Solver.Delayed.t -> (k,d,i) t -> k -> d Create.t -> unit =
    fun t dem k events ->
      mark_dem t dem k;
    (** record waiters *)
      let iter ev =
      Debug.dprintf2 debug "[Demon] @[Attach event %a@]@\n"
        Create.print ev;
        match ev with
        | Create.EventDom (cl,dom,data) ->
          Solver.Delayed.attach_dom t cl dom dem.dk_id (k,data)
        | Create.EventChange (cl,data) ->
          Solver.Delayed.attach_cl t cl dem.dk_id (k,data)
        | Create.EventRegCl (cl,data) ->
          Solver.Delayed.attach_reg_cl t cl dem.dk_id (k,data)
        | Create.EventRegSem (sem,data) ->
          Solver.Delayed.attach_reg_sem t sem dem.dk_id (k,data)
      in
      List.iter iter events


  type ('k,'i) state =
  | SUnborn
  | SAlive of 'i
  | SDead
  | SRedirected of 'k

  let is_attached (type k) (type d) (type i) t (dem: (k,d,i) t) (k:k) =
    let module DemTable = (val (Solver.Delayed.get_env t dem.dk_data)) in
    match DemTable.Key.M.find_opt k DemTable.state with
    | None -> SUnborn
    | Some (Alive(_,i)) -> SAlive i
    | Some Dead -> SDead
    | Some (Redirected k') -> SRedirected k'

  exception NotAlive

  let set_info (type k) (type d) (type i) t (dem: (k,d,i) t) (k:k) (i:i)  =
    let module DemTable = (val (Solver.Delayed.get_env t dem.dk_data)) in
    match DemTable.Key.M.find_exn NotAlive k DemTable.state with
    | Alive(w,_) ->
      let module DemTable' = struct
        include DemTable
        let state = DemTable.Key.M.add k (Alive(w,i)) DemTable.state
      end
      in
      Solver.Delayed.set_env t dem.dk_data (module DemTable')
    | _ -> raise NotAlive


  exception CantBeKilled

  let kill (type k) (type d) (type i) t (dem: (k,d,i) t) (k:k) =
    try
      let module DemTable = (val (Solver.Delayed.get_env t dem.dk_data)) in
      Debug.dprintf4 debug "[Demon] @[Kill dem %a %a@]@\n"
        Dem.print dem.dk_id DemTable.Key.print k;
      let module DemTable' = struct
        include DemTable
        let state = DemTable.Key.M.change (function
          | Some Dead -> raise Exit
          | _ -> Some Dead)
          k DemTable.state
      end in
      Solver.Delayed.set_env t dem.dk_data (module DemTable')
    with Exit -> ()

end

module Fast = struct

  type 'd t = {
    dk_id : ('d, unit) Types.dem;
    dk_data : 'd Solver.Events.Fired.event list Types.env;
    (** for throttling *)
    mutable dk_remaining: int; (** 0 if the demon is not the current one *)
    dk_current : 'd Solver.Events.Fired.event Queue.t; (** empty if idem *)
  }

  let create name = {
    dk_id   = Dem.create_key name;
    dk_data = Env.create_key name;
    dk_remaining = 0;
    dk_current = Queue.create ();
  }

  module type S = sig

    module Data: sig
      type t
      val print: t Pp.printer
    end

    val key: Data.t t

    (** never killed *)
    val immediate: bool
    val throttle: int (** todo int ref? *)
    (** number of time run in a row *)
    val wakeup: Solver.Delayed.t -> Data.t Solver.Events.Fired.event -> unit

  end


  module Register(D:S) = struct

    let run d () =
      assert (D.key.dk_remaining == 0);
      assert (Queue.is_empty D.key.dk_current);
      let rec last_rev q n = function
        | [] -> [],n
        | a::l ->
          let rem,n = last_rev q n l in
          if n > 0 then begin
            assert (rem == []);
            Queue.add a q;
            rem,(n-1)
          end
          else a::rem, n in
      let events = Solver.Delayed.get_env d D.key.dk_data in
      let events,n = last_rev D.key.dk_current D.throttle events in
      D.key.dk_remaining <- n;
      Solver.Delayed.set_env d D.key.dk_data events;
      let new_runable = if events != [] then Some () else None in
      let rec run_one () =
        if not (Queue.is_empty D.key.dk_current) then
          let event = Queue.pop D.key.dk_current in
          Debug.dprintf6 debug
            "[Demon] @[Run daemon fast %a:@\n@[%a@ %a@]@]@\n"
            Dem.print D.key.dk_id Solver.Events.Fired.print event
            D.Data.print (Solver.Events.Fired.get_data event);
          D.wakeup d event;
          Debug.dprintf0 debug "[Demon] @[Done@]@\n";
          if not D.immediate then Solver.Delayed.flush d;
          run_one () in
      try
        run_one ();
        assert (D.key.dk_remaining >= 0);
        assert (Queue.is_empty D.key.dk_current);
        D.key.dk_remaining <- 0;
        new_runable
      with exn -> (** Normally Contradiction *)
        assert (D.key.dk_remaining >= 0);
        D.key.dk_remaining <- 0;
        Queue.clear D.key.dk_current;
        raise exn

    let enqueue d event =
      assert (D.key.dk_remaining >= 0);
      if D.key.dk_remaining = 0 then
        let events = Solver.Ro.get_env d D.key.dk_data in
        Debug.dprintf4 debug
          "[Demon] @[schedule %a for %a@]@\n"
          Dem.print D.key.dk_id Solver.Events.Fired.print event;
        Solver.Ro.set_env d D.key.dk_data (event::events);
        if events = [] then Solver.EnqRun () else Solver.EnqAlready
      else begin
        Debug.dprintf4 debug
          "[Demon] @[schedule %a for %a now@]@\n"
          Dem.print D.key.dk_id Solver.Events.Fired.print event;
        Queue.add event D.key.dk_current;
        D.key.dk_remaining <- D.key.dk_remaining - 1;
        assert (D.key.dk_remaining >= 0);
        Solver.EnqAlready
      end


    let () =
      let print_demtable fmt d =
        Pp.print_list Pp.comma Solver.Events.Fired.print fmt d
      in
      Solver.register_env print_demtable D.key.dk_data;
    (** Interface for generic daemon *)
    let module Dem = struct
      type runable = unit
      let print_runable = Stdlib.DUnit.print
      let run = run

      type event = D.Data.t
      let print_event = D.Data.print
      let enqueue = enqueue

      let key = D.key.dk_id
      let immediate = D.immediate
    end in
    let module RDem = Solver.RegisterDem(Dem) in
    ()

    let init d =
      Solver.Delayed.set_env d D.key.dk_data [];

  end

  let attach d dem events =
    let open Create in
    List.iter (function
        | EventDom      (cl,dom,data) ->
          Solver.Delayed.attach_dom d cl dom dem.dk_id data
        | EventRegCl  (cl,data) ->
          Solver.Delayed.attach_reg_cl d cl dem.dk_id data
        | EventChange   (cl,data) ->
          Solver.Delayed.attach_cl d cl dem.dk_id data
        | EventRegSem (sem,data) ->
          Solver.Delayed.attach_reg_sem d sem dem.dk_id data) events

  let fresh_with_reg_cl dem s ty data =
    Cl.fresh ~to_reg:(dem.dk_id,data) s ty

end
