(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)
open Popop
open Types
open Stdlib
open Why_ptree
exception Not_supported of Loc.position

exception Unbound of (Loc.position * string)


type benv = {
  ctr : Ty.Constr.t DStr.H.t;
  term: Types.Cl.t  DStr.H.t;
}

let rec sort benv = function
  | PPTreal -> Arith.real
  | PPTint -> Arith.real
  | PPTbool -> Bool.ty
  | PPTvarid (ctr,loc) ->
    let ctr = DStr.H.find_exn benv.ctr (Unbound(Loc.extract loc,ctr)) ctr in
    Ty.ctr ctr
  | PPTexternal (args,ctr,loc) ->
    let l = List.map (sort benv) args in
    let ctr = DStr.H.find_exn benv.ctr (Unbound(Loc.extract loc,ctr)) ctr in
    Ty.app ctr (IArray.of_list l)
  | _ -> raise (Not_supported Loc.dummy_position)

let result_sort benv = function
  | PPredicate l -> (List.length l), Bool.ty
  | PFunction (l,result) -> (List.length l), sort benv result

let bind benv s cl =
  DStr.H.add benv.term s cl

let unbind benv s =
  DStr.H.remove benv.term s

let rec append_op env benv opc t acc =
  match t.pp_desc with
  | PPinfix (t1,op,t2) when op = opc  ->
    let l = append_op env benv opc t1 acc in
    let l = append_op env benv opc t2 l  in
    l
  | _ -> (of_lexpr env benv t)::acc

and of_lexpr env benv t =
  match t.pp_desc with
  | PPvar s -> DStr.H.find_exn benv.term (Unbound(Loc.extract t.pp_loc,s)) s
  | PPapp (s,l) ->
    let f = DStr.H.find_exn benv.term (Unbound(Loc.extract t.pp_loc,s)) s in
    let l = List.map (fun e -> of_lexpr env benv e) l in
    Uninterp.app_fun f l
  | PPconst (ConstInt s) ->
    Arith.cst (Q.of_string s)
  | PPconst (ConstReal q) ->
    Arith.cst q
  | PPinfix (t1,PPand,t2) ->
    let l = append_op env benv PPand t1 [] in
    let l = append_op env benv PPand t2 l in
    Bool._and            l
  | PPinfix (t1,PPor,t2) ->
    let l = append_op env benv PPor t1 [] in
    let l = append_op env benv PPor t2 l in
    Bool._or             l
  | PPinfix (t1,op,t2) ->
    let cl1,cl2 = Shuffle.seq2 (of_lexpr env benv) (t1,t2) in
    begin match op with
    | PPimplies -> Bool._or             [Bool._not cl1;cl2]
    | PPeq
    | PPiff     -> Equality.equality    [cl1;cl2]
    | PPneq     -> Equality.disequality [cl1;cl2]
    | PPadd     -> Arith.add            cl1 cl2
    | PPsub     -> Arith.sub            cl1 cl2
    | PPmul     -> Arith.mult           cl1 cl2
    | _ -> raise (Not_supported (Loc.extract t.pp_loc))
    end
  | PPprefix (op,e) ->
    let cl = of_lexpr env benv e in
    begin match op with
    | PPneg -> Arith.mult_cst (Q.of_int (-1)) cl
    | PPnot -> Bool._not cl
    end
  | PPforall (bl,ty,_,e) ->
    let result = sort benv ty in
    List.iter (fun name ->
        let cl = Variable.fresh result name in
        bind benv name cl) bl;
    let cl = of_lexpr env benv e in
    List.iter (fun name -> unbind benv name) bl;
    cl
  | PPforall_named (bl,ty,trs,e) ->
    let bl = List.map fst bl in
    of_lexpr env benv {t with pp_desc = PPforall (bl, ty, trs, e)}
  | PPconst (ConstTrue) -> Bool._true
  | PPconst (ConstFalse) -> Bool._false
  | _ -> raise (Not_supported (Loc.extract t.pp_loc))

let rec of_decl sched benv = function
  | [] -> raise (Not_supported Loc.dummy_position)
  | TypeDecl (_,_,name,Abstract)::l ->
    let ctr = Ty.Constr.create name in
    DStr.H.add benv.ctr name ctr;
    of_decl sched benv l
  | TypeDecl (loc,_,_,_)::_ ->
    raise (Not_supported (Loc.extract loc))
  | Logic (_,Symbols.Other,names,result)::l ->
    let arity, result = result_sort benv result in
    List.iter (fun (name,_) ->
        let cl = if arity > 0 then
            Uninterp.fresh_fun ~result
              ~arity name
          else
            Variable.fresh result name
        in
        DStr.H.add benv.term name cl) names;
    Scheduler.Scheduler.flush_delayed sched;
    of_decl sched benv l
  | Logic _::_ -> raise (Not_supported Loc.dummy_position)
  | Axiom(_,_,_,e)::l ->
    let envd = Scheduler.Scheduler.get_delayed sched in
    let cl = of_lexpr envd benv e in
    Solver.Delayed.register envd cl;
    Bool.set_true envd Explanation.pexpfact cl;
    Scheduler.Scheduler.flush_delayed sched;
    of_decl sched benv l
  | Goal(_,_,e)::_ ->
    let envd = Scheduler.Scheduler.get_delayed sched in
    let cl = of_lexpr envd benv e in
    Solver.Delayed.register envd cl;
    Bool.set_false envd Explanation.pexpfact cl;
    Scheduler.Scheduler.flush_delayed sched
  | _::l -> of_decl sched benv l


type result =
| Valid
| Idontknow
| Sat

let check_goal l =
  let sched = Scheduler.new_env [Variable.th_register;
                                 Uninterp.th_register; Arith.th_register;
                                 Bool.th_register;
                                 Equality.th_register] () in
  try
    let benv = {term = DStr.H.create 100; ctr = DStr.H.create 10} in
    of_decl sched benv l;
    Scheduler.Scheduler.stop_delayed sched;
    Scheduler.Scheduler.run_inf_step sched;
    Idontknow
  with Scheduler.Scheduler.Contradiction ->
    Valid

let read_file s =
  let cin = open_in s in
  let lb = Lexing.from_channel cin in
  Loc.set_file s lb;
  let _, decls = Loc.with_location (Why_parser.file Why_lexer.token) lb in
  decls

let read_split s =
  let cin = open_in s in
  let lb = Lexing.from_channel cin in
  Loc.set_file s lb;
  Loc.with_location (Why_parser.split_file Why_lexer.token) lb



let () = Exn_printer.register (fun fmt exn ->
  match exn with
  | Not_supported pos ->
    Format.fprintf fmt
      "%a@ popop can't convert alt-ergo construct"
    Loc.report_position pos
  | _ -> raise exn)
