(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

{
open Popop
type vars = Types.Cl.t array

let init_vars nb_var : vars =
  let a = Array.make nb_var Bool._true in
  for i = 0 to nb_var - 1 do
    a.(i) <- Variable.fresh Bool.ty (Printf.sprintf "x%i" (i+1))
  done;
  a


let get_lit vars i =
  let i = int_of_string i in
  let i,b = abs i - 1, i<0 in
  vars.(i),b

exception SyntaxError of string

let syntax_error s = raise (SyntaxError s)

let () = Exn_printer.register (fun fmt e -> match e with
  | SyntaxError s -> Format.pp_print_string fmt s
  | _ -> raise e)

}

let newline = '\n'
let space = [' ' '\t' '\r']+
let digit = ['0'-'9']
let sign = '-' | '+'
let integer = sign? digit+

rule find_header = parse
| newline  { Lexing.new_line lexbuf; find_header lexbuf }
| space    { find_header lexbuf }
| 'p'
    space+ "cnf"
    space+ (digit+ as nb_var)
    space+ (digit+ as nb_cls) { int_of_string nb_var,
                                int_of_string nb_cls }
| 'c' [^'\n']* '\n'  { Lexing.new_line lexbuf; find_header lexbuf }
| _
      { syntax_error "Can't find header" }

and clause vars acc = parse
| newline  { Lexing.new_line lexbuf; clause vars acc lexbuf }
| space { clause vars acc lexbuf }
| '0'  { List.rev acc }
| integer as i { clause vars ((get_lit vars i)::acc) lexbuf }
| _ { syntax_error "Bad clause" }

and file sched vars = parse
| newline  { Lexing.new_line lexbuf; file sched vars lexbuf }
| space { file sched vars lexbuf }
| '0'  { file sched vars lexbuf }
| integer as i { let lits = clause vars [get_lit vars i] lexbuf in
                 let d = Scheduler.Scheduler.get_delayed sched in
                 let cl = Bool.gen false lits in
                 Solver.Delayed.register d cl;
                 let pexp = Solver.Delayed.mk_pexp d Explanation.expfact () in
                 Bool.set_true d pexp cl;
                 Scheduler.Scheduler.flush_delayed sched;
                 file sched vars lexbuf
               }
| 'c' [^'\n']* ('\n' | eof)  { Lexing.new_line lexbuf; file sched vars lexbuf }
| eof { () }
| _ { syntax_error "Bad clauses" }

{

type answer =
| Sat
| Unsat
let parse sched s =
  let cin = open_in s in
  let lb = Lexing.from_channel cin in
  Loc.set_file s lb;
  let res =
    Loc.with_location (fun lexbuf ->
      let nb_vars, _ = find_header lexbuf in
      let vars = init_vars nb_vars in
      try
        file sched vars lexbuf;
        Scheduler.Scheduler.stop_delayed sched;
        Sat
      with Scheduler.Scheduler.Contradiction ->
        Unsat
    ) lb in
  if res = Unsat then raise Scheduler.Scheduler.Contradiction



let check_file filename =
  let sched = Scheduler.new_env [Variable.th_register;
                                 Bool.th_register_alone] () in
  try
    parse sched filename;
    Scheduler.Scheduler.run_inf_step sched;
    Sat
  with Scheduler.Scheduler.Contradiction -> Unsat


}

