(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)
open Popop
open Types
open Stdlib
open Smtlib2_ast

let debug_rename = Debug.register_flag
    ~desc:"for renaming the debug name of class"
    "rename"

type benv = {
  ctr : Ty.Constr.t DStr.H.t;
  term: Types.Cl.t  DStr.H.t;
}

exception Not_supported of Loc.position

exception Unbound of (Loc.position * string)

(** TODO check arity? *)
let declare_sort benv name _arity =
  let ctr = Ty.Constr.create name in
  DStr.H.add benv.ctr name ctr

let rec sort benv = function
  | SortIdentifier (_,IdSymbol(_,Symbol(loc,id))) ->
    let ctr = DStr.H.find_exn benv.ctr (Unbound(loc,id)) id in
    Ty.ctr ctr
  | SortIdSortMulti (_,IdSymbol(_,Symbol(loc,id)),(_,l)) ->
    let l = List.map (sort benv) l in
    let ctr = DStr.H.find_exn benv.ctr (Unbound(loc,id)) id in
    Ty.app ctr (IArray.of_list l)
  | SortIdentifier (loc,_)
  | SortIdSortMulti (loc,_,_) -> raise (Not_supported loc)

(** TODO: check that it is not a builtin construction? or allows to hide them *)
let declare_fun benv name sorts result =
  let arity = (List.length sorts) in
  let result = sort benv result in
  let cl = if arity > 0 then
      Uninterp.fresh_fun ~result
        ~arity name
    else
      Variable.fresh result name
  in
  DStr.H.add benv.term name cl

let qualidentifier benv args = function
  (** = *)
  | QualIdentifier(loc,IdSymbol(_,Symbol(_,"=")),None) ->
    if List.length args != 2 then raise (Not_supported loc);
    Equality.equality (Shuffle.shufflel args)
  (** distinct *)
  | QualIdentifier(_,IdSymbol(_,Symbol(_,"xor")),None)
  | QualIdentifier(_,IdSymbol(_,Symbol(_,"distinct")),None) ->
    Equality.disequality (Shuffle.shufflel args)
  (** not *)
  | QualIdentifier(_,IdSymbol(_,Symbol(_,"not")),None) ->
    assert (List.length args == 1);
    Bool._not (List.hd args)
  (** or *)
  | QualIdentifier(_,IdSymbol(_,Symbol(_,"or")),None) ->
    Bool._or (Shuffle.shufflel args)
  (** and *)
  | QualIdentifier(_,IdSymbol(_,Symbol(_,"and")),None) ->
    Bool._and (Shuffle.shufflel args)
  (** false *)
  | QualIdentifier(_,IdSymbol(_,Symbol(_,"false")),None) ->
    assert (args = []);
    Bool._false
  (** true *)
  | QualIdentifier(_,IdSymbol(_,Symbol(_,"true")),None) ->
    assert (args = []);
    Bool._true
  (** => *)
  | QualIdentifier(loc,IdSymbol(_,Symbol(_,"=>")),None) ->
    begin match args with
    | [b1;b2] ->  Bool.gen false [b1,true;b2,false]
    | _ -> raise (Not_supported loc)
    end
  (** ite *)
  | QualIdentifier(loc,IdSymbol(_,Symbol(_,"ite")),None) ->
    begin match args with
    | [cond;then_;else_] -> Equality.ite cond then_ else_
    | _ -> raise (Not_supported loc)
    end
  (** * *)
  | QualIdentifier(loc,IdSymbol(_,Symbol(_,"*")),None) ->
    begin match args with
    | [a1;a2] -> Arith.mult a1 a2
    | _ -> raise (Not_supported loc)
    end
  (** + *)
  | QualIdentifier(loc,IdSymbol(_,Symbol(_,"+")),None) ->
    begin match args with
    | [a1;a2] -> Arith.add a1 a2
    | _ -> raise (Not_supported loc)
    end
  (** - *)
  | QualIdentifier(loc,IdSymbol(_,Symbol(_,"-")),None) ->
    begin match args with
    | [a1] -> Arith.mult_cst Q.minus_one a1
    | [a1;a2] -> Arith.sub a1 a2
    | _ -> raise (Not_supported loc)
    end
  (** UserDefined *)
  | QualIdentifier(loc,IdSymbol(_,Symbol(_,name)),None) ->
    let cl = DStr.H.find_exn benv.term (Unbound(loc,name)) name in
    if args = [] then cl
    else Uninterp.app_fun cl args
  (** Not supported *)
  | QualIdentifier(loc,_,_) -> raise (Not_supported loc)


let unbind benv = function
  | VarBindingSymTerm (_, (Symbol(_,s)), _) ->
    DStr.H.remove benv.term s

let rec bind env benv = function
  | VarBindingSymTerm (_, (Symbol(_,s)), t) ->
    let cl = term env benv t in
    begin if Debug.test_flag debug_rename then
      let old_name = Pp.string_of Types.Cl.print cl in
      Types.Cl.rename cl s;
      Debug.dprintf3 debug_rename "[Smtlib2] @[rename cl: %s -> %a@]@\n"
        old_name Types.Cl.print cl
    end;
    DStr.H.add benv.term s cl

and term env benv = function
  | TermQualIdentifier (_,qualid) ->
    qualidentifier benv [] qualid
  | TermQualIdTerm (_,f,(_,args)) ->
    let args = List.map (fun e -> term env benv e) args in
    qualidentifier benv args f
  | TermLetTerm (_,(_,bl),t) ->
    List.iter (fun b -> bind env benv b) bl;
    let cl = term env benv t in
    List.iter (unbind benv) bl;
    cl
  | TermSpecConst (_,SpecConstsDec(_,q)) ->
    Arith.cst q
  | TermSpecConst (_,SpecConstNum(_,z)) ->
    Arith.cst (Q.of_bigint z)
  | TermForAllTerm (loc,_,_) | TermExistsTerm (loc,_,_)
  | TermSpecConst (loc,_) | TermExclimationPt (loc, _, _) ->
      raise (Not_supported loc)


let of_decl sched benv = function
  (** Noop *)
  | CSetInfo _ | CSetOption _ -> ()
  (** Supported *)
  | CSetLogic (_,Symbol(_,"QF_UF")) ->
    let env = Scheduler.Scheduler.get_delayed sched in
    Uninterp.th_register env
  | CSetLogic (_,Symbol(_,"QF_LRA")) ->
    let env = Scheduler.Scheduler.get_delayed sched in
    Arith.th_register env;
    DStr.H.add benv.ctr "Real" Arith.real_ctr
  | CSetLogic (_,Symbol(_,"QF_UFLRA")) ->
    let env = Scheduler.Scheduler.get_delayed sched in
    Uninterp.th_register env;
    Arith.th_register env;
    DStr.H.add benv.ctr "Real" Arith.real_ctr
  | CSetLogic (loc,_) -> raise (Not_supported loc)
  | CDeclareSort(_,Symbol(_,name),arity) ->
    declare_sort benv name arity
  | CDeclareFun(_,Symbol(_,s),(_,sorts),result) ->
    declare_fun benv s sorts result;
    Scheduler.Scheduler.flush_delayed sched
  | CAssert(_,t) ->
    let env = Scheduler.Scheduler.get_delayed sched in
    let cl = term env benv t in
    Solver.Delayed.register env cl;
    Bool.set_true env Explanation.pexpfact cl;
    Scheduler.Scheduler.flush_delayed sched
  | CCheckSat _ | CExit(_) -> () (** TODO better *)
  (** Not supported *)
  | CDefineSort (loc,_,_,_) | CDefineFun  (loc,_,_,_,_) | CPush (loc,_)
  | CPop (loc,_) | CGetAssert(loc) | CGetProof(loc) | CGetUnsatCore(loc)
  | CGetValue(loc,_) | CGetAssign(loc) | CGetOption(loc,_) | CGetInfo(loc,_)
    ->  raise (Not_supported loc)


type status =
  | Sat
  | Unsat
  | Unknown

let print_status fmt = function
  | Sat -> Format.pp_print_string fmt "sat"
  | Unsat -> Format.pp_print_string fmt "unsat"
  | Unknown -> Format.pp_print_string fmt "unknown"

let check_file l =
  let sched = Scheduler.new_env [Variable.th_register;
                                 Bool.th_register;
                                 Equality.th_register] () in
  try
    let benv = {term = DStr.H.create 100; ctr = DStr.H.create 10} in
    DStr.H.add benv.ctr "Bool" Bool.ty_ctr;
    List.iter (fun e -> of_decl sched benv e) l;
    Scheduler.Scheduler.stop_delayed sched;
    Scheduler.Scheduler.run_inf_step sched;
    Sat
  with Scheduler.Scheduler.Contradiction ->
    Unsat

let read_file s =
  let cin = open_in s in
  let lb = Lexing.from_channel cin in
  Loc.set_file s lb;
  Loc.with_location (Smtlib2_parser.commands Smtlib2_lexer.token) lb


let () = Exn_printer.register (fun fmt exn ->
  match exn with
  | Not_supported pos ->
    Format.fprintf fmt
    "%a@ popop can't convert smtlib2 construct"
    Loc.report_position pos
  | Unbound (pos,s) ->
    Format.fprintf fmt
    "%a@ unbound identifier %s"
    Loc.report_position pos s
  | _ -> raise exn)
