(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)

open Popop

module S = Solver

let debug = Debug.register_info_flag
  ~desc:"for the scheduler in the simple version"
  "sched_queue"

let debug_pushpop = Debug.register_info_flag
  ~desc:"for the scheduler push/pop"
  "sched_pushpop"

let var_decay = 1. /. 0.95

let stats_propa = Debug.register_stats_int ~name:"Scheduler.daemon" ~init:0
let stats_dec = Debug.register_stats_int ~name:"Scheduler.decision" ~init:0
let stats_con = Debug.register_stats_int ~name:"Scheduler.conflict" ~init:0

exception NeedStopDelayed

module Att = struct
  type t =
    | Daemon   of int * S.daemon_key
    | Decision of int * Explanation.chogen
  type prio = float
  type db = float Conflict.ChoGenH.t

  let get_prio db chogen =
    Conflict.ChoGenH.find_def db 0. chogen

  let le (x:t) (xp:float) (y:t) (yp:float) =
    match x, y with
    | Daemon (x,_)  , Daemon (y,_)   -> x <= y
    | Decision (x,_), Decision (y,_) ->
      if xp = yp then x <= y else xp >= yp (** min *)
    | Daemon _  , Decision _ -> true
    | Decision _, Daemon _   -> false
  let reprio db = function
    | Daemon _ -> 0.
    | Decision (_,chogen) -> get_prio db chogen
end
exception Contradiction

module Prio = Leftistheap.Make(Att)
let is_dec = function
  | Att.Daemon _ -> false
  | Att.Decision _ -> true

type pre =
  { pre_wakeup_daemons    : Prio.t;
    pre_prev_scheduler_state : pre option;
    pre_solver_state              : S.t;
    pre_learnt : Conflict.finalized list; (* bags *)
    pre_lastdec : Explanation.chogen;
  }

type t =
  { mutable wakeup_daemons    : Prio.t;
    mutable prev_scheduler_state : pre option;
    mutable solver_state      : S.t;
    mutable delayed           : S.Delayed.t option;
    mutable learnt : Conflict.finalized list; (* bags *)
    (* global *)
    decprio : Att.db;
    var_inc  : float ref;
  }
(** To treat in the reverse order *)

let get_t t = t.solver_state

let print_level fmt t =
  Format.fprintf fmt "%a (%i)"
    Explanation.Age.print (S.current_age t.solver_state)
    (S.current_nbdec t.solver_state)

let new_handler t =
  if t.delayed <> None then raise NeedStopDelayed;
  {wakeup_daemons    = t.wakeup_daemons;
   prev_scheduler_state = t.prev_scheduler_state;
   solver_state      = S.new_handler t.solver_state;
   learnt = t.learnt;
   delayed           = None;
   decprio = t.decprio;
   var_inc = t.var_inc
  }

let new_t t =
  { wakeup_daemons = Prio.empty;
    prev_scheduler_state = None;
    solver_state = S.new_handler t;
    learnt = [];
    delayed    = None;
    decprio = Conflict.ChoGenH.create 100;
    var_inc = ref 1.;
  }

let new_solver () = new_t (S.new_t ())

let push t chogen =
  Debug.dprintf0 debug_pushpop "[Scheduler] push@\n";
  let prev =
    { pre_wakeup_daemons    = t.wakeup_daemons;
      pre_prev_scheduler_state = t.prev_scheduler_state;
      pre_solver_state      = S.new_handler t.solver_state;
      pre_learnt = t.learnt;
      pre_lastdec = chogen;
    } in
  t.prev_scheduler_state <- Some prev;
  t.learnt <- []

let update_prio t chogen =
  Conflict.ChoGenH.change (function
      | None -> Some (!(t.var_inc))
      | Some i -> Some (i +. (!(t.var_inc)))) t.decprio chogen

let new_delayed =
  let daemon_count = ref (-1) in
  let dec_count = ref (-1) in
  fun t ->
    let sched_daemon att =
      incr daemon_count;
      Debug.dprintf1 debug "[Scheduler] New possible daemon:%i@\n"
        !daemon_count;
      t.wakeup_daemons <-
        Prio.insert t.decprio (Att.Daemon (!daemon_count,att))
          t.wakeup_daemons in
    let sched_decision dec =
      incr dec_count;
      Debug.dprintf1 debug "[Scheduler] New possible decisions prio:%i@\n"
        !dec_count;
      t.wakeup_daemons <- Prio.insert t.decprio (Att.Decision (!dec_count,dec))
          t.wakeup_daemons in
    S.new_delayed ~sched_daemon ~sched_decision t.solver_state

(*
  let rec apply_learnt llearnt t d =
    match llearnt with
    | [] -> d
    | a::l ->
      (** the first one should be the last conflict found so decide on it *)
      try
        let {Conflict.fin_dec = Explanation.GCho(cho,k)} = a d in
        S.flush d;
        List.iter (fun f -> ignore (f d); S.flush d) l;
        run_until_dec t d;
        run_dec t d t.wakeup_daemons
          (fun d dec -> Conflict.make_decision d dec cho k)
      with S.Contradiction pexp ->
        Debug.dprintf0 debug "[Scheduler] Contradiction during apply learnt@\n";
        conflict_analysis t pexp
*)

type choice =
  | Choice: ('k,'d) Explanation.cho * 'k * 'd * float -> choice
  | NoChoice

let rec apply_learnt (learntdec: Conflict.finalized) tags llearnt t d =
  try
    Debug.dprintf0 debug "[Scheduler] Apply previously learnt@\n";
    let iter_learnt f =
      Debug.dprintf2 debug "[Scheduler] @[Apply %a@]@\n"
        Conflict.print_finalized f;
      let clauses = f#conflict_add d in
      let cl = Bool.mk_clause clauses in
      (** What is learnt is a general consequence *)
      let pexp = Solver.Delayed.mk_pexp ~age:Explanation.Age.min
          d Conflict.explearnt (Conflict.ExpLearnt tags) in
      Solver.Delayed.register d cl;
      Bool.set_true d pexp cl;
      S.flush d in
    List.iter iter_learnt llearnt;
    iter_learnt learntdec;
    run_until_dec t d;
    Debug.dprintf0 debug "[Scheduler] Apply learntdec@\n";
    (** Do directly a decision on the last conflict learnt *)
    let fold_decisions acc cho' k' d' =
      match acc with
      | NoChoice ->
        let f' = Att.get_prio t.decprio (Explanation.GCho(cho',k')) in
        Choice(cho',k',d',f')
      | Choice(_,_,_,f) ->
        let f' = Att.get_prio t.decprio (Explanation.GCho(cho',k')) in
        if f >= f' then acc else Choice(cho',k',d',f')
    in
    match learntdec#decide {Conflict.fold_decisions} d NoChoice with
    | NoChoice ->
      d
    | Choice(cho,k,v,_) ->
      run_dec t d cho k v
  with S.Contradiction pexp ->
    Debug.dprintf0 debug "[Scheduler] Contradiction during apply learnt@\n";
    conflict_analysis t pexp

and pop_to t prev =
  Debug.dprintf2 debug_pushpop "[Scheduler] pop %a@\n"
    print_level t;
  t.wakeup_daemons <- prev.pre_wakeup_daemons;
  t.prev_scheduler_state <- prev.pre_prev_scheduler_state;
  t.solver_state <- prev.pre_solver_state;
  t.learnt <- prev.pre_learnt;
  new_delayed t

(*
  and conflict_analysis t pexp =
    Debug.incr stats_con;
    let learnt,_tags, _decs = Conflict.analyse t.solver_state pexp in
    let learnt,maxage = match learnt with
      | None -> raise Contradiction
      | Some learntmaxage -> learntmaxage in
    let rec rewind llearnt maxage prevo =
      match prevo with
      | None -> raise Contradiction
      | Some prev when
          Explanation.current_age
            (S.get_trail prev.pre_solver_state) <= maxage ->
        llearnt,prev
      | Some prev ->
        let llearnt = List.rev_append llearnt prev.pre_learnt in
        rewind llearnt maxage prev.pre_prev_scheduler_state
    in
    let llearnt,prev =
      rewind (learnt::t.learnt) maxage t.prev_scheduler_state in
    Debug.dprintf2 debug_pushpop "[Scheduler] Pop to level %a@\n"
      Explanation.Age.print maxage;
    pop_to llearnt t prev
*)

and conflict_analysis t pexp =
  Debug.incr stats_con;
  if Solver.current_nbdec t.solver_state = 0 then raise Contradiction
  else
    let pre_solver_state = match t.prev_scheduler_state with
      | None -> (* level 0 *) raise Contradiction
      | Some prev -> prev.pre_solver_state in
    let learnt,tags, decs =
      Conflict.analyse pre_solver_state t.solver_state
        (Solver.get_trail t.solver_state) pexp in
    t.var_inc := !(t.var_inc) *. var_decay;
    Bag.iter (update_prio t) decs;
    let learnt = match learnt with
      | None -> raise Contradiction
      | Some learnt -> learnt in
    (** We look for the level where lastlearnt is not false *)
    let rec rewind t learnt llearnt prevo =
      match prevo with
      | None -> raise Contradiction
      | Some prev ->
        let prevage = Explanation.current_age (** for printing *)
            (S.get_trail prev.pre_solver_state) in
        let d = pop_to t prev in
        let llearnt_all = List.rev_append llearnt prev.pre_learnt in
        let open Conflict in
        match learnt#test d with
        | True  -> raise Types.Impossible (** understand why that happend *)
        | False -> rewind t learnt llearnt_all prev.pre_prev_scheduler_state
        | ToDecide ->
          (** we found the level *)
          Debug.dprintf4 debug_pushpop "[Scheduler] Pop to level %a %a@\n"
            Explanation.Age.print prevage print_level t;
          t.learnt <- learnt::llearnt_all;
          d,learnt,llearnt
    in
    let d,learntdec,llearnt =
      rewind t learnt t.learnt t.prev_scheduler_state in
    t.wakeup_daemons <- Prio.reprio t.decprio t.wakeup_daemons;
    apply_learnt learntdec tags llearnt t d

and try_run_dec:
  type k d. t -> S.Delayed.t -> Prio.t -> ((k,d) Explanation.cho) -> k ->
    S.Delayed.t = fun t d prio cho k ->
    (** First we verify its the decision is at this point needed *)
    try
      match Conflict.choose_decision d cho k with
      | Conflict.DecNo ->
        t.wakeup_daemons <- prio;
        d (** d can be precised by choose_decision *)
      | Conflict.DecTodo todo ->
        Debug.incr stats_dec;
        S.delayed_stop d;
        Debug.dprintf2 debug_pushpop
          "[Scheduler] Make decision: level %a@\n"
          print_level t;
        (** The registered state keep the old prio *)
        push t (Explanation.GCho(cho,k));
        (** We use the priority list without the decision only in the
            branch where the decision is made *)
        t.wakeup_daemons <- prio;
        let declevel = Explanation.new_dec (S.get_trail t.solver_state) in
        let d = new_delayed t in
        todo d declevel;
        d
    with S.Contradiction pexp ->
      Debug.dprintf0 debug "[Scheduler] Contradiction@\n";
      conflict_analysis t pexp

and run_dec:
  type k d. t -> S.Delayed.t -> ((k,d) Explanation.cho) -> k -> d ->
    S.Delayed.t = fun t d cho k v ->
    (** First we verify its the decision is at this point needed *)
    try
      Debug.incr stats_dec;
      S.delayed_stop d;
      Debug.dprintf2 debug_pushpop
        "[Scheduler] Make decision: level %a@\n"
        print_level t;
      (** The registered state keep the old prio *)
      push t (Explanation.GCho(cho,k));
      let declevel = Explanation.new_dec (S.get_trail t.solver_state) in
      let d = new_delayed t in
      Conflict.make_decision d cho k v declevel;
      d
    with S.Contradiction pexp ->
      Debug.dprintf0 debug "[Scheduler] Contradiction@\n";
      conflict_analysis t pexp

and run_until_dec t d =
  let act = Prio.min t.wakeup_daemons in
  match act with
  | Att.Daemon (_,att) -> begin
      let _, prio = Prio.extract_min t.wakeup_daemons in
      Debug.incr stats_propa;
      t.wakeup_daemons <- prio;
      S.run_daemon d att;
      S.flush d;
      run_until_dec t d
    end
  | Att.Decision (_,_) -> ()


let run_one_step t d =
  let act, prio = Prio.extract_min t.wakeup_daemons in
  match act with
  | Att.Daemon (_,att) -> begin
      Debug.incr stats_propa;
      t.wakeup_daemons <- prio;
      try
        S.run_daemon d att; d
      with S.Contradiction pexp ->
        Debug.dprintf0 debug "[Scheduler] Contradiction@\n";
        conflict_analysis t pexp
    end
  | Att.Decision (_,Explanation.GCho(cho,k)) -> try_run_dec t d prio cho k

let rec flush t d =
  try
    S.flush d; d
  with S.Contradiction pexp ->
    Debug.dprintf0 debug "[Scheduler] Contradiction@\n";
    let d = conflict_analysis t pexp in
    flush t d

let rec run_inf_step ~nodec t d =
  let d = flush t d in
  if not (Prio.is_empty t.wakeup_daemons)
  && not (nodec && is_dec (Prio.min t.wakeup_daemons))
  then
    let d = run_one_step t d in
    run_inf_step ~nodec t d
  else
    S.delayed_stop d

let run_inf_step ?(nodec=false) t =
  if t.delayed <> None then raise NeedStopDelayed;
  let d = new_delayed t in
  run_inf_step ~nodec t d

let get_delayed t =
  match t.delayed with
  | Some d -> d
  | None   ->
    let d = new_delayed t in
    t.delayed <- Some d;
    d

let flush_delayed t =
  match t.delayed with
  | None -> ()
  | Some d ->
    t.delayed <- Some (flush t d)

let stop_delayed t =
  match t.delayed with
  | None -> ()
  | Some d ->
    let d = flush t d in
    S.delayed_stop d;
    t.delayed <- None

let run_exn ?nodec ~theories f =
  let t = new_solver () in
  begin try
      let d = get_delayed t in
      List.iter (fun f -> f d) theories;
      Solver.flush d;
      f d
    with S.Contradiction _ ->
      Debug.dprintf0 debug
        "[Scheduler] Contradiction during initial assertion@\n";
      raise Contradiction
  end;
  stop_delayed t;
  run_inf_step ~nodec:(nodec=Some ()) t;
  get_delayed t

let run ?nodec ~theories f =
  try
    `Done (run_exn ?nodec ~theories f)
  with Contradiction ->
    `Contradiction
