(**************************************************************************)
(*                                                                        *)
(*  This file is part of Frama-C.                                         *)
(*                                                                        *)
(*  Copyright (C) 2013                                                    *)
(*    CEA (Commissariat à l'énergie atomique et aux énergies              *)
(*         alternatives)                                                  *)
(*                                                                        *)
(*  you can redistribute it and/or modify it under the terms of the GNU   *)
(*  Lesser General Public License as published by the Free Software       *)
(*  Foundation, version 2.1.                                              *)
(*                                                                        *)
(*  It is distributed in the hope that it will be useful,                 *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU Lesser General Public License for more details.                   *)
(*                                                                        *)
(*  See the GNU Lesser General Public License version 2.1                 *)
(*  for more details (enclosed in the file licenses/LGPLv2.1).            *)
(*                                                                        *)
(**************************************************************************)
open Popop
open Stdlib
open Types
open Solver

let debug = Debug.register_info_flag
  ~desc:"for the uninterpreted function theory"
  "uninterp"

type t = App of Cl.t * Cl.t

let sem : t sem = Sem.create_key "UF"

module Th = struct
  module T = struct
    type r = t
    type t = r
    let equal n m =
      match n, m with
      | App (g1,f1), App(g2,f2) -> Cl.equal g1 g2 && Cl.equal f1 f2

    let hash n =
      match n with
      | App(g,f) -> 3 * (Cl.hash g) + 5 * (Cl.hash f)

    let compare n m =
      match n, m with
      | App(g1,f1), App(g2,f2) ->
        let c = Cl.compare g1 g2 in
        if c <> 0 then c
        else Cl.compare f1 f2

    let print fmt v =
      match v with
      | App (f,g) -> Format.fprintf fmt "(@[%a@],@,@[%a@])"
        Cl.print  f Cl.print g

  end

  include MkDatatype(T)

  let key = sem

    (* let norm d s = *)
    (*   Delayed.propagate d (Delayed.index d sem s) *)

    (* let propagate ~propagate s = *)
    (*   match s with *)
    (*   | App(f,g) -> propagate f; propagate g *)

end

let print = Th.print

module ThE = RegisterSem(Th)

let funty_ctr = Ty.Constr.create "Fun"
let funty = Ty.ctr funty_ctr

let partty_ctr = Ty.Constr.create "PartialFun"
let partty = Ty.ctr partty_ctr

let app f g ty = Cl.index sem (App(f,g)) ty

let appl f l ty =
  let rec aux = function
    | [] -> assert false
    | [a] -> a
    | a::l -> app a (aux l) partty in
  let l = aux l in
  app f l ty

let result_of_sort = Ty.H.create 20

let register_sort ?dec ty =
  Ty.H.add result_of_sort ty dec

(** Bool can't register itself it is linked before uninterp *)
let () = register_sort Bool.ty ~dec:Bool.make_dec

let result_to_dec = Cl.H.create 20

let fresh_fun ~result ~arity name =
  assert (arity > 0);
  let cl = Cl.fresh name funty in
  Cl.H.add result_to_dec cl (result,arity);
  cl

let app_fun cl args =
  let (sort,arity) = Cl.H.find_exn result_to_dec Impossible cl in
  assert (List.length args = arity);
  let appcl = appl cl args sort in
  appcl

let fun1 ty s =
  let f = fresh_fun ~result:ty ~arity:1 s in
  (fun x -> app_fun f [x])

let fun2 ty s =
  let f = fresh_fun ~result:ty ~arity:2 s in
  (fun x1 x2 -> app_fun f [x1;x2])

let fun3 ty s =
  let f = fresh_fun ~result:ty ~arity:3 s in
  (fun x1 x2 x3 -> app_fun f [x1;x2;x3])

let fun4 ty s =
  let f = fresh_fun ~result:ty ~arity:4 s in
  (fun x1 x2 x3 x4 -> app_fun f [x1;x2;x3;x4])

let fun5 ty s =
  let f = fresh_fun ~result:ty ~arity:5 s in
  (fun x1 x2 x3 x4 x5 -> app_fun f [x1;x2;x3;x4;x5])


type expsubst = ThE.t
let expsubst : expsubst Explanation.exp =
  Explanation.Exp.create_key "Uninterp"

module DaemonPropa = struct
  type k = ThE.t
  let key = Demon.Key.create "Uninterp.DaemonPropa"

  module Key = ThE
  module Data = DUnit
  type info = unit let default = ()

  let is_unborn d v =
    let open Demon.Key in
    match is_attached d key v with
    | SDead | SRedirected _ | SAlive () -> false
    | SUnborn -> true

  let attach d f g v =
    let open Demon.Create in
    assert (is_unborn d v);
    Demon.Key.attach d key v
      [EventChange(f,()); EventChange(g,())]

  let immediate = true (** can be false *)
  let wakeup d (clsem:k) _ev () =
    match ThE.sem clsem with
    | App(f,g) as v ->
      Debug.dprintf4 debug "[Uninterp] @[wakeup own %a v:%a@]@\n"
        Cl.print (ThE.cl clsem) Th.print v;
      let v' = App(Delayed.find d f, Delayed.find d g) in
      assert (not (Th.equal v v'));
      let pexp = Delayed.mk_pexp d expsubst clsem in
      let clsem' = ThE.index v' (ThE.ty clsem) in
      Delayed.set_sem d pexp (ThE.cl clsem) (ThE.clsem clsem');
      Demon.AliveRedirected clsem'
end

module RDaemonPropa = Demon.Key.Register(DaemonPropa)

module DaemonInit = struct
  let key = Demon.Key.create "Uninterp.DaemonInit"

  module Key = DUnit
  module Data = DUnit
  type info = unit let default = ()

  let immediate = false

  let wakeup d () ev () =
    List.iter
      (function Events.Fired.EventRegSem(clsem,()) ->
        begin
          let clsem = ThE.coerce_clsem clsem in
          let v = ThE.sem clsem in
          let own = ThE.cl clsem in
          Debug.dprintf2 debug "[Uninterp] @[init %a@]@\n" Th.print v;
          let dec = Ty.H.find_def result_of_sort None (Cl.ty own) in
          Opt.iter (fun dec -> Variable.add_dec ~dec d own) dec;
          if DaemonPropa.is_unborn d clsem then
          match v with
          | App(f,g) ->
            Delayed.register d f; Delayed.register d g;
            let f' = Delayed.find d f in
            let g' = Delayed.find d g in
            if Cl.equal f' f && Cl.equal g' g then
              DaemonPropa.attach d f g clsem
            else
              let v' = App(f',g') in
              let pexp = Delayed.mk_pexp d expsubst clsem in
              let clsem = ThE.clsem (ThE.index v' (Cl.ty own)) in
              Delayed.set_sem d pexp own clsem
        end
      | _ -> raise UnwaitedEvent
      ) ev;
    Demon.AliveReattached


end

module RDaemonInit = Demon.Key.Register(DaemonInit)

let th_register env =
  RDaemonPropa.init env;
  RDaemonInit.init env;
  Demon.Key.attach env
    DaemonInit.key () [Demon.Create.EventRegSem(sem,())];

module ExpSubst = struct
  open Conflict
  (* open IterExp *)
  open ComputeConflict

  type t = expsubst

  let print fmt clsem =
    match ThE.sem clsem with
    | App(f,g) ->
      Format.fprintf fmt "Subst(%a,%a)" Cl.print f Cl.print g

(*
  let iterexp t age = function
    | App(f,g) ->
      need_sem t age sem (App(f,g));
      need_cl_repr t age f;
      need_cl_repr t age g
*)
  let analyse :
    type a. Conflict.ComputeConflict.t ->
    Explanation.age -> a Explanation.con -> t -> a Conflict.rescon =
    fun t age con clsem ->
      begin match ThE.sem clsem with
      | App(f,g) ->
        let fty = Cl.ty f in
        let gty = Cl.ty g in
        Equality.GenEquality.equality t age f (get_repr_at t age f) fty;
        Equality.GenEquality.equality t age g (get_repr_at t age g) gty
      end;
      return_nothing con

  let expdom _ _ _ _ _ _ = raise Impossible (** don't propa dom *)

  let key = expsubst

end

module EExpSubst = Conflict.RegisterExp(ExpSubst)
