open OUnit

open Tests_lib
open Scheduler

let new_env =
  new_env [Variable.th_register; Uninterp.th_register; Arith.th_register]

let register sched cl =
  let d = Scheduler.get_delayed sched in
  Solver.Delayed.register d cl;
  Scheduler.flush_delayed sched

let merge sched cl1 cl2 =
  let d = Scheduler.get_delayed sched in
  Solver.Delayed.merge d Explanation.pexpfact cl1 cl2;
  Scheduler.flush_delayed sched

let is_equal sched cl1 cl2 =
  Scheduler.stop_delayed sched;
  Scheduler.run_inf_step sched;
  let env = Scheduler.get_t sched in
  Solver.is_equal env cl1 cl2


let solve0 () =
  let env = new_env () in
  let a  = Variable.cst Arith.real "ar" in
  let b  = Variable.cst Arith.real "br" in
  let _1 = Arith.cst Q.one in
  let a1 = Arith.add a _1 in
  let b1 = Arith.add b _1 in
  register env a1; register env b1;
  merge env a1 b1;
  assert_bool "a+1 = b+1 => a = b" (is_equal env a b)

let solve1 () =
  let env = new_env () in
  let a,b  = Shuffle.seq2 (Variable.cst Arith.real) ("ar","br") in
  let _1 = Arith.cst Q.one in
  let a1 = Arith.add a _1 in
  let b1 = Arith.add b _1 in
  let _2 = Arith.cst (Q.of_int 2) in
  let a2 = Arith.add a _2 in
  let b2 = Arith.add b _2 in
  Shuffle.seql' (register env) [a1; b1; a2; b2];
  merge env a1 b1;
  assert_bool "a+1 = b+1 => a+2 = b+2" (is_equal env a2 b2)

let solve2 () =
  let env = new_env () in
  let a,b  = Shuffle.seq2 (Variable.cst Arith.real) ("ar","br") in
  let _1 = Arith.cst Q.one in
  let a1 = Arith.add a _1 in
  let b1 = Arith.add b _1 in
  let _2 = Arith.cst (Q.of_int 2) in
  let a2 = Arith.add a _2 in
  let b2 = Arith.add b _2 in
  Shuffle.seql' (register env) [a1; b1; a2; b2];
  merge env a2 b1;
  assert_bool "a+2 = b+1 => a+1 = b" (is_equal env a1 b)

let solve3 () =
  let env = new_env () in
  let a,b  = Shuffle.seq2 (Variable.cst Arith.real) ("ar","br") in
  let _1 = Arith.cst Q.one in
  let b1 = Arith.add b _1 in
  let _2 = Arith.cst (Q.of_int 2) in
  let a2 = Arith.add a _2 in
  let _3 = Arith.cst (Q.of_int 3) in
   Shuffle.seql [
    (fun () ->
      Shuffle.seql' (register env) [b1;a2];
      merge env a2 b1;
    );
    (fun () ->
      Shuffle.seql' (register env) [a;_2];
      merge env a _2;
    );
    (fun () ->
      register env _3;
    );
   ];
  assert_bool "" (not (is_equal env b _2));
  assert_bool "a+2 = b+1 => a = 2 => b = 3" (is_equal env b _3)


let solve4 () =
  let env = new_env () in
  let a,b,c =
    Shuffle.seq3 (Variable.cst Arith.real) ("ar","br","cr") in
  let t1 = Arith.cst (Q.of_int 2) in
  let t1 = Arith.add t1 c in
  let t1 = Arith.add a t1  in
  let t1' = (Arith.cst (Q.of_int 1)) in
  let t1' = Arith.add b t1' in
  let t2  = a in
  let t2' = Arith.cst (Q.of_int 2) in
  let t2' = Arith.add t2' b in
  let t3' = Arith.cst (Q.of_int (-3)) in
  Shuffle.seql [
    (fun () ->
      Shuffle.seql' (register env) [t1;t1'];
      merge env t1 t1');
    (fun () ->
      Shuffle.seql' (register env) [t2;t2'];
      merge env t2 t2');
    (fun () -> register env t3');
  ];
  assert_bool "a+(2+c) = b+1 => a = 2 + b => c = -3" (is_equal env c t3')


let solve5 () =
  let env = new_env () in
  let a  = Variable.cst Arith.real "ar" in
  let b  = Variable.cst Arith.real "br" in
  let c  = Variable.cst Arith.real "cr" in
  let t1 = Arith.sub b c in
  let t1  = Arith.add a t1  in
  let t1' = (Arith.cst (Q.of_int 2)) in
  let t2  = a in
  let t2' = Arith.cst (Q.of_int 2) in
  let t3 = Arith.add b c in
  let t3' = Arith.add b b in
  Shuffle.seql [
    (fun () ->
      Shuffle.seql' (register env) [t1;t1'];
      merge env t1 t1');
    (fun () ->
      Shuffle.seql' (register env) [t2;t2'];
      merge env t2 t2');
    (fun () ->
      Shuffle.seql' (register env) [t3;t3'];)
  ];
  assert_bool "a+(b-c) = 2 => a = 2 => b + c = 2b" (is_equal env t3 t3')


let basic = "Arith.Basic" &: [solve0; solve1; solve2; solve3; solve4; solve5]

let mult0 () =
  let env = new_env () in
  let a  = Variable.cst Arith.real "ar" in
  let b  = Variable.cst Arith.real "br" in
  let t1  = Arith.sub a b  in
  let t1' = Arith.mult a b in
  let t2  = a in
  let t2' = Arith.cst (Q.of_int 1) in
  let t3 = Arith.mult_cst (Q.of_int 2) b in
  let t3' = Arith.cst (Q.of_int 1) in
  Shuffle.seql [
    (fun () ->
      Shuffle.seql' (register env) [t1;t1'];
      merge env t1 t1');
    (fun () ->
      Shuffle.seql' (register env) [t2;t2'];
      merge env t2 t2');
    (fun () ->
      Shuffle.seql' (register env) [t3;t3'];)
  ];
  assert_bool "a - b = a * b -> a = 1 -> 1 = 2b" (is_equal env t3 t3')

(** test that mult normalization trigger the needed solve *)
let mult1 () =
  let env = new_env () in
  let a  = Variable.cst Arith.real "ar" in
  let b  = Variable.cst Arith.real "br" in
  let c  = Variable.cst Arith.real "cr" in
  let t1  = Arith.mult a b  in
  let t1  = Arith.add a t1  in
  let t1' = Arith.add b c in
  let t1' = Arith.mult t1' a in
  let t2  = a in
  let t2' = Arith.cst (Q.of_int 2) in
  let t3 = c in
  let t3' = Arith.cst (Q.of_int 1) in
  Shuffle.seql [
    (fun () ->
      Shuffle.seql' (register env) [t1;t1'];
      merge env t1 t1');
    (fun () ->
      Shuffle.seql' (register env) [t2;t2'];
      merge env t2 t2');
    (fun () ->
      Shuffle.seql' (register env) [t3;t3'];)
  ];
  assert_bool "a + (a * b) = (b + c) * a -> a = 2 -> c = 1"
    (is_equal env t3 t3')

let mult = "Arith.Mult" &: [mult0;mult1]


let files = ["tests/tests_altergo_arith.split";
             "tests/tests_popop.split";
            ]

let altergo = TestList (List.map Tests_lib.test_split files)

let smtlib2sat =
  "smtlib2-lra-sat" >:::
    tests_smt2 Popop_of_smtlib2.Sat "tests/smtlib2/lra/sat/"

let smtlib2unsat =
  "smtlib2-lra-unsat" >:::
    tests_smt2 Popop_of_smtlib2.Unsat "tests/smtlib2/lra/unsat/"

let tests = TestList [basic;mult;altergo;smtlib2sat;smtlib2unsat]
