open OUnit

open Scheduler
open Tests_lib


let theories =  [Variable.th_register; Bv.th_register]

let run = Scheduler.run_exn ~nodec:() ~theories
let ($$) f x = f x

let rec propagate_all env = function
  | [] -> ()
  | x::rest -> register env x; propagate_all env rest

let _42 = Z.of_int 42
let _62 = Z.of_int 62
let _63 = Z.of_int 63

let tests_pretty_print () =
  (*
    let test_cst cst =
    assert_equal
    ~printer:(fun x -> x)
    ~msg:(Pp.sprintf "test_cst %a" Z.pp_print cst)
    (Z.format "[%b]" cst)
    (Pp.string_of_wnl Bv.D.print {Bv.aun=cst;Bv.azer=cst})
    in
    test_cst Z.zero;
    test_cst _62;
    test_cst _42;
  *)
  let test exp aun azer =
    assert_equal
      ~printer:(fun x -> x)
      exp
      (Pp.string_of_wnl Bv.D.print {Bv.aun;Bv.azer})
  in
  test "[0?]" Z.zero Z.one;
  test "[0]" Z.zero Z.zero ;
  test "[01]" Z.one Z.one ;
  test "[11]" Z.minus_one Z.minus_one;
  test "[?1]" Z.one Z.minus_one;
  test "[0??????]" Z.zero _63;
  test "[0XXXXX0]" _62 Z.zero

let simpleprint var env str =
  let xd = Bv.D.get_dom_bv  env Bv.dom var in
  Format.print_string str;
  Bv.D.print Format.std_formatter xd;
  Format.print_string "\n"

type bvtestexp =
|Tor of bvtestexp * bvtestexp
|Tand of  bvtestexp * bvtestexp
|Txor of  bvtestexp * bvtestexp
|Tnot of bvtestexp
|Tsl of bvtestexp * bvtestexp
|Tsr of bvtestexp * bvtestexp
|Tcst of int
|Tvar of string
|Tcl of Popop.Types.Cl.t

let exmp1 =
  let x = Tvar "x" in
  let y = Tvar "y" in
  [(Tand (x,y), Tcst 0);
   (x,Tnot (Tcst 0))]

let rec testexp_to_test (e: bvtestexp) env =
  match e with
  | Tor (b,c) ->
    let bb = testexp_to_test b env in
    let cc = testexp_to_test c env in
    propagate_all env [bb;cc];
    Bv.bor bb cc
  | Tand (b,c) ->
    let bb = testexp_to_test b env in
    let cc = testexp_to_test c env in
    propagate_all env [bb;cc];
    Bv.band bb cc
  | Txor (b,c) ->
    let bb = testexp_to_test b env in
    let cc = testexp_to_test c env in
    propagate_all env [bb;cc];
    Bv.bxor bb cc
  | Tsl (b,c) ->
    let bb = testexp_to_test b env in
    let cc = testexp_to_test c env in
    propagate_all env [bb;cc];
    Bv.bsl bb cc
  | Tsr(b,c) ->
    let bb = testexp_to_test b env in
    let cc = testexp_to_test c env in
    propagate_all env [bb;cc];
    Bv.bsr bb cc
  | Tnot c ->
    let cc = testexp_to_test c env in
    register env cc;
    Bv.bnot cc
  | Tcst n -> Bv.cnst n
  | Tvar s -> Variable.fresh Bv.ty s
  | Tcl c -> c

let rec constraint_list el env =
  match el with
  | [] ->   ()
  | (b,c) :: rest ->
    let bb = testexp_to_test b env in
    let cc =  testexp_to_test c env in
    propagate_all env [bb;cc];
    merge env bb cc;
    assert_bool "inside" (is_equal   env bb cc);
    constraint_list rest env

let simple_tests num () =
  let x = Variable.fresh  Bv.ty  "x" in
  let y = Variable.fresh  Bv.ty  "y" in
  let t = Variable.fresh  Bv.ty  "t" in
  let z = Variable.fresh  Bv.ty  "z" in
  let r = Variable.fresh  Bv.ty  "r" in
  let cl_one = Bv.cnst (-1) in
  let cl_zero = Bv.cnst 0 in
  let initial env =  propagate_all env [cl_one; cl_zero;x;y;t;z;r]
  in
  match num with
  | 1 -> (* x = Cst 1 *)
    let env = run $$ fun env ->
      initial env;
      merge env x cl_one
    in
    assert_bool "x = 1 " (is_equal  env x cl_one)

  | 2 ->  (* x|y = 0 *)
    let lside = Bv.bor x y in
    let env = run $$ fun env ->
      initial env;
      propagate_all env [lside];
      merge env lside cl_zero
    in
    assert_bool "x | y = 0" (is_equal  env lside cl_zero);
    assert_bool "x = 0" (is_equal  env x cl_zero);
    assert_bool "y = 0" (is_equal  env y cl_zero)

  | 3 ->  (* x & y = 1 *)
    let lside = Bv.band x y in
    let env = run $$ fun env ->
      initial env;propagate_all env [lside];
      merge env lside cl_one
    in
    assert_bool "x & y = 1" (is_equal  env lside cl_one);
    assert_bool "x = 1" (is_equal  env x cl_one);
    assert_bool "y = 1" (is_equal  env y cl_one)

  | 4 ->  (* x & 1 = 0  *)
    let lside = Bv.band x cl_one in
    let env = run $$ fun env ->
      initial env;
      propagate_all env [lside];
      merge env lside cl_zero;
    in
    assert_bool "x & 1 = 0" (is_equal env lside cl_zero);
    assert_bool "x = 0" (is_equal  env x cl_zero)

  | 5 -> (* x | 0 = 1  *)
    let lside = Bv.bor x cl_zero in
    let env = run $$ fun env ->
      initial env;
      propagate_all env [lside];
      merge env lside cl_one
    in
    assert_bool "x | 0 = 1" (is_equal env lside cl_one);
    assert_bool "x = 1" (is_equal env x cl_one)

  | 6 -> (* x = not 0 *)
    let rside = Bv.bnot cl_zero in
    let env = run $$ fun env ->
      initial env;
      register env rside;
      register env cl_zero;
      register env x;
      merge env x rside
    in
    assert_bool "x = ¬ 0" (is_equal env rside x)

  | 7 ->
    let _ = run $$ fun env ->
      initial env;
      constraint_list exmp1 env
    in
    ()
  | 8 ->
    let exmp2 =
      [(Tand (Tcl x, Tcl y), Tcst 0);
       (Tcl x,Tnot (Tcst 0))] in
    let env = run $$ fun env ->
      initial env;
      constraint_list exmp2 env
    in
    assert_bool "x = 1" (is_equal env x cl_one);
    assert_bool "y = 0" (is_equal env y cl_zero)

  | 9 ->
    (*
      a-  z ⊕ 101 = (x & y) | (t & y)
      b-  y ⊕ x = 110
      c-  t & x = 100
      d-  z ⊕ t = 010
      e-   t = 100
    *)
    let exmp =
      let x = Tcl x in
      let y = Tcl y in
      let z = Tcl z in
      let t = Tcl t in
      let zxor5 = Txor (z,Tcst 7) in
      let xandy = Tand (x,y) in
      let tandy = Tand (t,y) in
      let yorx = Tor (y, x) in
      let tandx = Tand (t,x) in
      let zxort = Txor (z,t) in
      [(zxor5, Tor (xandy,tandy));
       (yorx, Tcst 7);
       (tandx, Tcst 4);
       (zxort,Tcst 3);
       (t, Tcst 4);
       (x, Tcst 4)
      ] in
    let env = run $$ fun env ->
      initial env;
      constraint_list exmp env;
    in
    assert_bool "x = 100" (is_equal env x (Bv.cnst 4));
    assert_bool "y = 011" (is_equal env y (Bv.cnst 3));
    assert_bool "z = 100" (is_equal env z (Bv.cnst 7));
    assert_bool "t = 100" (is_equal env t (Bv.cnst 4));
  (* simpleprint x env "\nx="; *)
  (* simpleprint y env "y="; *)
  (* simpleprint z env "z="; *)
  (* simpleprint t env "t=" *)

  | 10 -> (*  100 >> 2 = r  *)
    let exmp = [ Tsr (Tcst 4, Tcst 2), Tcl r] in
    let env = run $$ fun env ->
      initial env;
      constraint_list exmp env
    in
    assert_bool "r = 01" (is_equal env r (Bv.cnst 1));

  | 11 -> (* x << 1 = 1010 ==> x = ?101*)
    let exmp =
      [(Tsl (Tcl x, Tcst 1), Tcst 10);
       (Tand (Tcl x, Tcl cl_one), Tcst 5)] in
    let env = run $$ fun env ->
      initial env;
      constraint_list exmp env
    in
    assert_bool "x = 101" (is_equal env x (Bv.cnst 5));

  | 12 -> (* x >> 1 = 1010 ==> x = 01010?*)
    let exmp =
      [(Tsr (Tcl x, Tcst 1), Tcst 10);
       (Tor (Tcl x, Tcl cl_zero), Tcst 20)] in
    let env = run $$ fun env ->
      initial env;
      constraint_list exmp env
    in
    assert_bool "x = 10100" (is_equal env x (Bv.cnst 20));

  | _ -> failwith "No test by that number in tests_bv.ml"

let rec gen_list seed op =
  match (op seed) with
  | None  -> []
  | Some x -> seed :: (gen_list x op)

let incr until seed =
  if seed > until then None
  else Some (seed + 1)

let rec gen_tests = function
  | [] -> []
  | x :: rest ->
    (("Number " ^ (string_of_int x)) >::
	(simple_tests x)) :: gen_tests rest

let basic = "Bv.Basic" >::: ((
  "pretty print" >:: tests_pretty_print) :: (gen_tests (gen_list 1 (incr 12))) )

let tests = TestList[basic]
