open OUnit

let size = 10
let size0 = size * 10

module Trace : sig
  type t
  type r

  val mk: unit -> t

  val int: t -> int -> unit
  val bool: t -> bool -> unit

  val result: t -> r
  val equal: r -> r -> bool
end = struct
  type t = Cryptokit.hash
  type r = string

  let mk = Cryptokit.Hash.sha1

  let int (t:t) i =
    let int32 t i =
      let add t i = t#add_byte (i land 255); i lsr 8 in
      add t (add t (add t (add t i))) in
    match Sys.word_size with
    | 32 -> ignore (int32 t i)
    | 64 -> ignore (int32 t (int32 t i))
    | _ -> assert false (** absurd: No other possible achitecture supported *)

  let bool t b = if b then t#add_byte 1 else t#add_byte 0

  let result t = t#result
  let equal = Stdlib.DStr.equal

end

module Patricia = Stdlib.DInt.M
module Binary   = Extmap.Make(Stdlib.DInt)
module OCaml    = Map.Make(Stdlib.DInt)

(** utils *)
let same_result = function
  | [] -> true
  | a::l ->
    List.for_all (fun b -> Trace.equal a b) l

let run_trace f rnd genmap =
  let tr = Trace.mk () in
  let rnd = Random.State.copy rnd in
  (** todo choose different gen *)
  let iter k d = Trace.int tr k; Trace.int tr d in
  let fold acc k d = Trace.int tr acc; iter k d; (acc + 1) in
  let () = f ~iter ~fold ~rnd tr genmap in
  Trace.result tr

let (!) rnd = (Random.State.int rnd 99999) - 9999

let compare_trace f l =
  let rnd = Random.State.make [|Shuffle.int 1000|] in
  same_result (List.map (run_trace f rnd) l)

let adds ~rnd ~add n empty =
  Util.foldi (fun m _ -> let x = !rnd in add x x m) empty 0 n

let rms ~rnd ~rm n map =
  Util.foldi (fun m _ -> rm (!rnd) m) map 0 n

(** Generic tests *)
type genmap = (module Map.S with type key = int)

let genmaps : genmap list = [(module Patricia);(module Binary);(module OCaml)]

let empty ~iter ~fold:_ ~rnd:_ _tr (gm : genmap) =
  let module M = (val gm) in
  M.iter iter M.empty

let singleton ~iter ~fold:_ ~rnd _tr (gm : genmap) =
  let module M = (val gm) in
  M.iter iter (M.singleton (!rnd) (!rnd))

let iter ~iter ~fold:_ ~rnd _tr (gm : genmap) =
  let module M = (val gm) in
  M.iter iter (adds ~rnd ~add:M.add size0 M.empty)

let fold ~iter:_ ~fold ~rnd tr (gm : genmap) =
  let module M = (val gm) in
  Trace.int tr (M.fold fold (adds ~rnd ~add:M.add size0 M.empty) 0)

let remove ~iter ~fold:_ ~rnd _tr (gm : genmap) =
  let module M = (val gm) in
  let rnd' = Random.State.copy rnd in
  let m = adds ~rnd  ~add:M.add    size0 M.empty in
  M.iter iter m;
  let m = rms  ~rnd:rnd' ~rm:M.remove  size  m in
  M.iter iter m

let merge ~iter ~fold:_ ~rnd _tr (gm : genmap) =
  let module M = (val gm) in
  let m1 = adds ~rnd ~add:M.add size0 M.empty in
  let m2 = adds ~rnd ~add:M.add size0 M.empty in
  let merge k d1 d2 =
    let d = k * (Opt.get_def 1 d1) * (Opt.get_def 1 d2) in
    if d mod 2 = 0 then None else Some (d / 2) in
  M.iter iter (M.merge merge m1 m2)

let test f ()  =
  assert_bool "compare" (compare_trace f genmaps)

let gen = ["empty" >:: test empty;
           "singleton" >:: test singleton;
           "iter" >:: test iter;
           "fold" >:: test fold;
           "remove" >:: test remove;
           "merge" >:: test merge;
          ]

(** Extended tests *)
type 'a pint = int
module type MapII = sig
  type 'a data
  include Map_intf.Map with type key = int and type 'a data := 'a pint
end
type extmap = (module MapII)



module Int = struct
  include Stdlib.DInt
  let tag x = x
 end

module Pat = Intmap.Make(Int)
module HPat = Pat.Make(Int)


let extmaps : extmap list = [(module Patricia);(module Binary);(module HPat)]

let empty ~iter ~fold:_ ~rnd:_ tr (gm : extmap) =
  let module M = (val gm) in
  assert_bool "check_invariant" (M.check_invariant M.empty);
  M.iter iter M.empty

let singleton ~iter ~fold:_ ~rnd _tr (gm : extmap) =
  let module M = (val gm) in
  let m = M.singleton (!rnd) (!rnd) in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m

let iter ~iter ~fold:_ ~rnd _tr (gm : extmap) =
  let module M = (val gm) in
  let m = adds ~rnd ~add:M.add size0 M.empty in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m

let fold ~iter:_ ~fold ~rnd tr (gm : extmap) =
  let module M = (val gm) in
  let m = adds ~rnd ~add:M.add size0 M.empty in
  assert_bool "check_invariant" (M.check_invariant m);
  Trace.int tr (M.fold fold m 0)

let remove ~iter ~fold:_ ~rnd _tr (gm : extmap) =
  let module M = (val gm) in
  let rnd' = Random.State.copy rnd in
  let m = adds ~rnd  ~add:M.add    size0 M.empty in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m;
  let m = rms  ~rnd:rnd' ~rm:M.remove  size  m in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m

let merge ~iter ~fold:_ ~rnd _tr (gm : extmap) =
  let module M = (val gm) in
  let m1 = adds ~rnd ~add:M.add size0 M.empty in
  let m2 = adds ~rnd ~add:M.add size0 M.empty in
  let merge k d1 d2 =
    let d = k * (Opt.get_def 1 d1) * (Opt.get_def 1 d2) in
    if d mod 2 = 0 then None else Some (d / 2) in
  let m = M.merge merge m1 m2 in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m


let union ~iter ~fold:_ ~rnd _tr (gm : extmap) =
  let module M = (val gm) in
  let m1 = adds ~rnd ~add:M.add size0 M.empty in
  let m2 = adds ~rnd ~add:M.add size M.empty in
  let m3 = adds ~rnd ~add:M.add size0 M.empty in
  let m1' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m1 m2 in
  let m2' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m3 m2 in
  let union k d1 d2 =
    let d = k * d1 * d2 in
    if d mod 2 = 0 then None else Some (d / 2) in
  let m = M.union union m1' m2' in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m

let inter ~iter ~fold:_ ~rnd _tr (gm : extmap) =
  let module M = (val gm) in
  let m1 = adds ~rnd ~add:M.add size0 M.empty in
  let m2 = adds ~rnd ~add:M.add size M.empty in
  let m3 = adds ~rnd ~add:M.add size0 M.empty in
  let m1' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m1 m2 in
  let m2' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m3 m2 in
  let inter k d1 d2 =
    let d = k * d1 * d2 in
    if d mod 2 = 0 then None else Some (d / 2) in
  let m = M.inter inter m1' m2' in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m


let union_merge ~iter ~fold:_ ~rnd _tr (gm : extmap) =
  let module M = (val gm) in
  let m1 = adds ~rnd ~add:M.add size0 M.empty in
  let m2 = adds ~rnd ~add:M.add size M.empty in
  let m3 = adds ~rnd ~add:M.add size0 M.empty in
  let m1' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m1 m2 in
  let m2' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m3 m2 in
  let union k d1 d2 =
    let d = k * (Opt.get_def 1 d1) * d2 in
    if d mod 2 = 0 then None else Some (d / 2) in
  let m = M.union_merge union m1' m2' in
  assert_bool "check_invariant" (M.check_invariant m);
  M.iter iter m


let disjoint ~iter:_ ~fold:_ ~rnd tr (gm : extmap) =
  let module M = (val gm) in
  let m1 = adds ~rnd ~add:M.add (size0 * 10) M.empty in
  let m2 = adds ~rnd ~add:M.add (size * 10) M.empty in
  let m3 = adds ~rnd ~add:M.add (size0 * 10) M.empty in
  let m1' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m1 m2 in
  let m2' = M.union (fun _ d1 d2 -> Some (d1 + d2)) m3 m2 in
  let disjoint _ d1 d2 = not ((d1 - d2) mod (size0 * 10) = 0) in
  Trace.bool tr (M.disjoint disjoint m1' m2');
  let disjoint _ _ _ = false in
  Trace.bool tr (M.disjoint disjoint m1' m2')

let disjointe ~iter:_ ~fold:_ ~rnd tr (gm : extmap) =
  let module M = (val gm) in
  Trace.bool tr (M.set_disjoint M.empty M.empty)

let test f ()  =
  assert_bool "compare" (compare_trace f extmaps)

let ext = ["empty" >:: test empty;
           "singleton" >:: test singleton;
           "iter" >:: test iter;
           "fold" >:: test fold;
           "remove" >:: test remove;
           "merge" >:: test merge;
           "union" >:: test union;
           "inter" >:: test inter;
           "union_merge" >:: test union_merge;
           "disjoint" >:: test disjoint;
           "disjointe" >:: test disjointe;
          ]

let tests = "Map" >::: gen@ext
